"""
===================================================================================
Reconstruction with Robust and Unbiased Model-BAsed Spherical Deconvolution (RUMBA)
===================================================================================

This example shows how to use RUMBA-SD to reconstruct fiber orientation
density functions (fODFs). This model was introduced by
:footcite:t:`CanalesRodriguez2015`.

RUMBA-SD uses a priori information about the fiber response function (axial
and perpendicular diffusivities) to generate a convolution kernel mapping the
fODFs on a sphere to the recorded data. The fODFs are then estimated using an
iterative, maximum likelihood estimation algorithm adapted from Richardson-Lucy
(RL) deconvolution :footcite:p:`Richardson1972`. Specifically, the RL algorithm
assumes Gaussian noise, while RUMBA assumes Rician/Noncentral Chi noise -- these
more accurately reflect the noise generated by MRI scanners
:footcite:p:`Constantinides1997`. This algorithm also contains an optional
compartment for estimating an isotropic volume fraction to account for partial
volume effects. RUMBA-SD works with single- and multi-shell data, as well as
data recorded in Cartesian or spherical coordinate systems.

The result from RUMBA-SD can be smoothed by applying total variation spatial
regularization (termed RUMBA-SD + TV), a technique which promotes a more
coherent estimate of the fODFs across neighboring voxels :footcite:p:`Rudin1992`.
This regularization ability is also included in this implementation.

This example will showcase how to:
    1. Estimate the fiber response function
    2. Reconstruct the fODFs voxel-wise or globally with TV regularization
    3. Visualize fODF maps

To begin, we will load the data, consisting of 10 b0s and 150 non-b0s with a
b-value of 2000.
"""

import matplotlib.pyplot as plt
import numpy as np

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames, get_sphere
from dipy.direction import peak_directions, peaks_from_model
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
from dipy.reconst.csdeconv import auto_response_ssst, recursive_response
from dipy.reconst.rumba import RumbaSDModel
from dipy.segment.mask import median_otsu
from dipy.sims.voxel import single_tensor_odf
from dipy.viz import actor, window

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")
data, affine = load_nifti(hardi_fname)

bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

sphere = get_sphere(name="symmetric362")

###############################################################################
# Step 1. Estimation of the fiber response function
# =================================================
#
# There are multiple ways to estimate the fiber response function.
#
# **Strategy 1: use default values**
# One simple approach is to use the values included as the default arguments
# in the RumbaSDModel constructor. The white matter response, `wm_response`
# has three values corresponding to the tensor eigenvalues
# (1.7e-3, 0.2e-3, 0.2e-3). The model has compartments for cerebrospinal
# fluid (CSF) (`csf_response`) and grey matter (GM) (`gm_response`) as well,
# with these mean diffusivities set to 3.0e-3 and 0.8e-3 respectively
# :footcite:p:`CanalesRodriguez2015` These default values will often be adequate
# as RUMBA-SD is robust against impulse response imprecision
# :footcite:p:`DellAcqua2007`.

rumba = RumbaSDModel(gtab)
print(
    f"wm_response: {rumba.wm_response}, "
    + f"csf_response: {rumba.csf_response}, "
    + f"gm_response: {rumba.gm_response}"
)

###############################################################################
# We can visualize what this default response looks like.

# Enables/disables interactive visualization
interactive = False

scene = window.Scene()

evals = rumba.wm_response
evecs = np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]]).T

response_odf = single_tensor_odf(sphere.vertices, evals=evals, evecs=evecs)
# Transform our data from 1D to 4D
response_odf = response_odf[None, None, None, :]
response_actor = actor.odf_slicer(response_odf, sphere=sphere, colormap="plasma")

scene.add(response_actor)
print("Saving illustration as default_response.png")
window.record(scene=scene, out_path="default_response.png", size=(200, 200))

if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Default response function.

scene.rm(response_actor)

###############################################################################
# **Strategy 2: estimate from local brain region**
# The `csdeconv` module contains functions for estimating this response.
# `auto_response_sst` extracts an ROI in the center of the brain and isolates
# single fiber populations from the corpus callosum using an FA mask with a
# threshold of 0.7. These voxels are used to estimate the response function.

response, _ = auto_response_ssst(gtab, data, roi_radii=10, fa_thr=0.7)
print(response)

###############################################################################
# This response contains the estimated eigenvalues in its first element, and
# the estimated S0 in the second. The eigenvalues are all we care about for
# using RUMBA-SD.
#
# We can visualize this estimated response as well.

evals = response[0]
evecs = np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]]).T

response_odf = single_tensor_odf(sphere.vertices, evals=evals, evecs=evecs)
# transform our data from 1D to 4D
response_odf = response_odf[None, None, None, :]
response_actor = actor.odf_slicer(response_odf, sphere=sphere, colormap="plasma")
scene.add(response_actor)
print("Saving illustration as estimated_response.png")
window.record(scene=scene, out_path="estimated_response.png", size=(200, 200))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Estimated response function.

scene.rm(response_actor)

###############################################################################
# **Strategy 3: recursive, data-driven estimation**
# The other method for extracting a response function uses a recursive
# approach. Here, we initialize a "fat" response function, which is used in
# CSD. From this deconvolution, the voxels with one peak are extracted and
# their data is averaged to get a new response function. This is repeated
# iteratively until convergence :footcite:p:`Tax2014`.
#
# To shorten computation time, a mask can be estimated for the data.

b0_mask, mask = median_otsu(data, median_radius=2, numpass=1, vol_idx=np.arange(10))

rec_response = recursive_response(
    gtab,
    data,
    mask=mask,
    sh_order_max=8,
    peak_thr=0.01,
    init_fa=0.08,
    init_trace=0.0021,
    iter=4,
    convergence=0.001,
    parallel=True,
    num_processes=2,
)

###############################################################################
# We can now visualize this response, which will look like a pancake.

rec_response_signal = rec_response.on_sphere(sphere)
# transform our data from 1D to 4D
rec_response_signal = rec_response_signal[None, None, None, :]
response_actor = actor.odf_slicer(rec_response_signal, sphere=sphere, colormap="plasma")

scene.add(response_actor)
print("Saving illustration as recursive_response.png")
window.record(scene=scene, out_path="recursive_response.png", size=(200, 200))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Recursive response function.

scene.rm(response_actor)

###############################################################################
# Step 2. fODF Reconstruction
# ===========================
#
# We will now use the estimated response function with the RUMBA-SD model to
# reconstruct the fODFs. We will use the default value for `csf_response` and
# `gm_response`. If one doesn't wish to fit these compartments, one can specify
# either argument as `None`. This will result in the corresponding volume
# fraction map being all zeroes. The GM compartment can only be accurately
# estimated with at least 3-shell data. With less shells, it is recommended
# to only keep the compartment for CSF. Since this data is single-shell, we
# will only compute the CSF compartment.
#
# RUMBA-SD can fit the data voxelwise or globally. By default, a voxelwise
# approach is used (`voxelwise` is set to `True`). However, by setting
# `voxelwise` to false, the whole brain can be fit at once. In this global
# setting, one can specify the use of TV regularization with `use_tv`, and the
# model can log updates on its progress and estimated signal-to-noise ratios by
# setting `verbose` to True. By default, both `use_tv` and `verbose` are set to
# `False` as they have no bearing on the voxelwise fit.
#
# When constructing the RUMBA-SD model, one can also specify `n_iter`,
# `recon_type`, `n_coils`, `R`, and `sphere`. `n_iter` is the number of
# iterations for the iterative estimation, and the default value of 600
# should be suitable for most applications. `recon_type` is the technique used
# by the MRI scanner to reconstruct the MRI signal, and should be either 'smf'
# for 'spatial matched filter', or 'sos' for 'sum-of-squares'; 'smf' is a
# common choice and is the default, but the specifications of the MRI scanner
# used to collect the data should be checked. If 'sos' is used, then it's
# important to specify `n_coils`, which is the number of coils in the MRI
# scanner. With 'smf', this isn't important and the default argument of 1 can
# be used. `R` is the acceleration factor of the MRI scanner, which is termed
# the iPAT factor for SIEMENS, the ASSET factor for GE, or the SENSE factor
# for PHILIPS. 1 is a common choice, and is the default for the model. This
# is only important when using TV regularization, which will be covered later
# in the tutorial. Finally, `sphere` specifies the sphere on which to construct
# the fODF. The default is 'repulsion724' sphere, but this tutorial will use
# `symmetric362`.

rumba = RumbaSDModel(gtab, wm_response=response[0], gm_response=None, sphere=sphere)

###############################################################################
# For efficiency, we will only fit a small part of the data. This is the same
# portion of data used in
# :ref:`sphx_glr_examples_built_reconstruction_reconst_csd.py`.

data_small = data[20:50, 55:85, 38:39]

###############################################################################
# **Option 1: voxel-wise fit**
# This is the default approach for generating ODFs, wherein each voxel is fit
# sequentially.
#
# We will estimate the fODFs using the 'symmetric362' sphere. This
# will take about a minute to compute.

rumba_fit = rumba.fit(data_small)
odf = rumba_fit.odf()

###############################################################################
# The inclusion of RUMBA-SD's CSF compartment means we can also extract
# the isotropic volume fraction map as well as the white matter volume
# fraction map (the fODF sum at each voxel). These values are normalized such
# that they sum to 1. If neither isotropic compartment is included, then the
# isotropic volume fraction map will all be zeroes.

f_iso = rumba_fit.f_iso
f_wm = rumba_fit.f_wm

###############################################################################
# We can visualize these maps using adjacent heatmaps.

fig, axs = plt.subplots(1, 2, figsize=(12, 5))

ax0 = axs[0].imshow(f_wm[..., 0].T, origin="lower")
axs[0].set_title("Voxelwise White Matter Volume Fraction")

ax1 = axs[1].imshow(f_iso[..., 0].T, origin="lower")
axs[1].set_title("Voxelwise Isotropic Volume Fraction")

plt.colorbar(ax0, ax=axs[0])
plt.colorbar(ax1, ax=axs[1])

plt.savefig("wm_iso_partition.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# White matter and isotropic volume fractions
#
#
#
# To visualize the fODFs, it's recommended to combine the fODF and the
# isotropic components. This is done using the `RumbaFit` object's method
# `combined_odf_iso`. To reach a proper scale for visualization, the argument
# `norm=True` is used in FURY's `odf_slicer` method.

combined = rumba_fit.combined_odf_iso

fodf_spheres = actor.odf_slicer(
    combined, sphere=sphere, norm=True, scale=0.5, colormap=None
)
scene.add(fodf_spheres)
print("Saving illustration as rumba_odfs.png")
window.record(scene=scene, out_path="rumba_odfs.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# RUMBA-SD fODFs

scene.rm(fodf_spheres)

###############################################################################
# We can extract the peaks from these fODFs using `peaks_from_model`. This
# will reconstruct the fODFs again, so will take about a minute to run.

rumba_peaks = peaks_from_model(
    model=rumba,
    data=data_small,
    sphere=sphere,
    relative_peak_threshold=0.5,
    min_separation_angle=25,
    normalize_peaks=False,
    parallel=True,
    num_processes=4,
)

###############################################################################
# For visualization, we scale up the peak values.

peak_values = np.clip(rumba_peaks.peak_values * 15, 0, 1)
peak_dirs = rumba_peaks.peak_dirs

fodf_peaks = actor.peak_slicer(peak_dirs, peaks_values=peak_values)
scene.add(fodf_peaks)

window.record(scene=scene, out_path="rumba_peaks.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# RUMBA-SD peaks

scene.rm(fodf_peaks)

###############################################################################
# **Option 2: global fit**
# Instead of the voxel-wise fit, RUMBA also comes with an implementation of
# global fitting where all voxels are fit simultaneously. This comes with some
# potential benefits such as:
#
# 1. More efficient fitting due to matrix parallelization, in exchange for
#    larger demands on RAM (>= 16 GB should be sufficient)
# 2. The option for spatial regularization; specifically, TV regularization is
#    built into the fitting function (RUMBA-SD + TV)
#
# This is done by setting `voxelwise` to `False`, and setting `use_tv` to
# `True`.
#
# TV regularization requires a volume without any singleton dimensions, so
# we'll have to start by expanding our data slice.

rumba = RumbaSDModel(
    gtab,
    wm_response=response[0],
    gm_response=None,
    voxelwise=False,
    use_tv=True,
    sphere=sphere,
)
data_tv = data[20:50, 55:85, 38:40]

###############################################################################
# Now, we fit the model in the same way. This will take about 90 seconds.

rumba_fit = rumba.fit(data_tv)

odf = rumba_fit.odf()
combined = rumba_fit.combined_odf_iso

###############################################################################
# Now we can visualize the combined fODF map as before.

fodf_spheres = actor.odf_slicer(
    combined, sphere=sphere, norm=True, scale=0.5, colormap=None
)

scene.add(fodf_spheres)
print("Saving illustration as rumba_global_odfs.png")
window.record(scene=scene, out_path="rumba_global_odfs.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# RUMBA-SD + TV fODFs
#
#
#
# This can be compared with the result without TV regularization, and one can
# observe that the coherence between neighboring voxels is improved.

scene.rm(fodf_spheres)

###############################################################################
# For peak detection, `peaks_from_model` cannot be used as it doesn't support
# global fitting approaches. Instead, we'll compute our peaks using a for loop.

shape = odf.shape[:3]
npeaks = 5  # maximum number of peaks returned for a given voxel
peak_dirs = np.zeros((shape + (npeaks, 3)))
peak_values = np.zeros((shape + (npeaks,)))

for idx in np.ndindex(shape):  # iterate through each voxel
    # Get peaks of odf
    direction, pk, _ = peak_directions(
        odf[idx], sphere, relative_peak_threshold=0.5, min_separation_angle=25
    )

    # Calculate peak metrics
    if pk.shape[0] != 0:
        n = min(npeaks, pk.shape[0])
        peak_dirs[idx][:n] = direction[:n]
        peak_values[idx][:n] = pk[:n]

# Scale up for visualization
peak_values = np.clip(peak_values * 15, 0, 1)

fodf_peaks = actor.peak_slicer(
    peak_dirs[:, :, 0:1, :], peaks_values=peak_values[:, :, 0:1, :]
)
scene.add(fodf_peaks)

print("Saving illustration as rumba_global_peaks.png")
window.record(scene=scene, out_path="rumba_global_peaks.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# RUMBA-SD + TV peaks

scene.rm(fodf_peaks)

###############################################################################
# References
# ----------
#
# .. footbibliography::
#



"""
========================================
Surface seeding for tractography
========================================

Surface seeding is a way to generate initial position for tractography
from cortical surfaces position :footcite:p:`StOnge2018`.
"""

from fury.io import load_polydata
from fury.utils import (
    get_actor_from_polydata,
    get_polydata_triangles,
    get_polydata_vertices,
    normals_from_v_f,
)
import numpy as np

from dipy.data import get_fnames
from dipy.tracking.mesh import (
    random_coordinates_from_surface,
    seeds_from_surface_coordinates,
)
from dipy.viz import actor, window

###############################################################################
# Fetch and load a surface

brain_lh = get_fnames(name="fury_surface")
polydata = load_polydata(brain_lh)

###############################################################################
# Extract the triangles and vertices

triangles = get_polydata_triangles(polydata)
vts = get_polydata_vertices(polydata)

###############################################################################
# Display the surface
# ===============================================
#
# First, create an actor from the polydata, to display in the scene

scene = window.Scene()
surface_actor = get_actor_from_polydata(polydata)

scene.add(surface_actor)
scene.set_camera(position=(-500, 0, 0), view_up=(0.0, 0.0, 1))

# Uncomment the line below to show to display the window
# window.show(scene, size=(600, 600), reset_camera=False)
window.record(scene=scene, out_path="surface_seed1.png", size=(600, 600))

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Initial cortical surface
#
#
# Generate a list of seeding positions
# ==============================================================
#
# Choose the number of seed

nb_seeds = 100000
nb_triangles = len(triangles)

###############################################################################
# Get a list of triangles indices and trilinear coordinates for each seed

tri_idx, trilin_co = random_coordinates_from_surface(nb_triangles, nb_seeds)

###############################################################################
# Get the 3d cartesian position from triangles indices and trilinear
# coordinates

seed_pts = seeds_from_surface_coordinates(triangles, vts, tri_idx, trilin_co)

###############################################################################
# Compute normal and get the normal direction for each seeds

normals = normals_from_v_f(vts, triangles)
seed_n = seeds_from_surface_coordinates(triangles, normals, tri_idx, trilin_co)

###############################################################################
# Create dot actor for seeds (blue)

seed_actors = actor.dot(seed_pts, colors=(0, 0, 1), dot_size=4.0)

###############################################################################
# Create line actors for seeds normals (green outside, red inside)

normal_length = 0.5
normal_in = np.tile(seed_pts[:, np.newaxis, :], (1, 2, 1))
normal_out = np.tile(seed_pts[:, np.newaxis, :], (1, 2, 1))
normal_in[:, 0] -= seed_n * normal_length
normal_out[:, 1] += seed_n * normal_length

normal_in_actor = actor.line(normal_in, colors=(1, 0, 0))
normal_out_actor = actor.line(normal_out, colors=(0, 1, 0))

###############################################################################
# Visualise seeds and normals along the surface

scene = window.Scene()
scene.add(surface_actor)
scene.add(seed_actors)
scene.add(normal_in_actor)
scene.add(normal_out_actor)
scene.set_camera(position=(-500, 0, 0), view_up=(0.0, 0.0, 1))

# Uncomment the line below to show to display the window
# window.show(scene, size=(600, 600), reset_camera=False)
window.record(scene=scene, out_path="surface_seed2.png", size=(600, 600))

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Surface seeds with normal orientation
#
#
# References
# ----------
#
# .. footbibliography::
#



"""
=============================================
Reconstruction of Bingham Functions from ODFs
=============================================

This example shows how to reconstruct Bingham functions from orientation
distribution functions (ODFs). Reconstructed Bingham functions can be
useful to quantify properties from ODFs such as fiber dispersion
:footcite:p:`Riffert2014`, :footcite:p:`NetoHenriques2018`.

To begin, let us import the relevant functions and load data consisting
of 10 b0s and 150 non-b0s with a b-value of 2000s/mm2.
"""

from dipy.core.gradients import gradient_table
from dipy.core.sphere import unit_icosahedron
from dipy.data import get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
from dipy.reconst.bingham import sf_to_bingham, sh_to_bingham
from dipy.reconst.csdeconv import ConstrainedSphericalDeconvModel, auto_response_ssst
from dipy.viz import actor, window
from dipy.viz.plotting import image_mosaic

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")
data, affine = load_nifti(hardi_fname)

bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

###############################################################################
# To properly fit Bingham functions, we recommend the use of a larger number of
# directions to sample the ODFs. For this, we load a `sphere` object with 12
# vertices sampling a 3D sphere (the icosahedron). We further subdivide the
# faces of this `sphere` representation five times, to get 10242 directions.

sphere = unit_icosahedron.subdivide(n=5)

nd = sphere.vertices.shape[0]
print("The number of directions on the sphere is {}".format(nd))

###############################################################################
# Step 1. ODF estimation
# ======================
#
# Before fitting Bingham functions, we must reconstruct ODFs. In this example,
# fiber ODFs (fODFs) will be reconstructed using the Constrained Spherical
# Deconvolution (CSD) method :footcite:p:`Tournier2007`. For simplicity, we
# will refer to fODFs as ODFs.
# In the main tutorial of CSD (see
# :ref:`sphx_glr_examples_built_reconstruction_reconst_csd.py`), several
# strategies to define the fiber response function are discussed. Here, for
# the sake of simplicity, we will use the response function estimates from a
# local brain region:

response, ratio = auto_response_ssst(gtab, data, roi_radii=10, fa_thr=0.7)

# Let us now compute the ODFs using this response function:

csd_model = ConstrainedSphericalDeconvModel(gtab, response, sh_order_max=8)

###############################################################################
# For efficiency, we will only fit a small part of the data.

data_small = data[20:50, 55:85, 38:39]
csd_fit = csd_model.fit(data_small)

###############################################################################
# Let us visualize the ODFs

csd_odf = csd_fit.odf(sphere)

interactive = False

scene = window.Scene()

fodf_spheres = actor.odf_slicer(
    csd_odf, sphere=sphere, scale=0.9, norm=False, colormap="plasma"
)
scene.add(fodf_spheres)

print("Saving the illustration as csd_odfs.png")
window.record(scene, out_path="csd_odfs.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Fiber ODFs (fODFs) reconstructed using Constrained Spherical
# Deconvolution (CSD). For simplicity, we will refer to them just as ODFs.
#
#
# Step 2. Bingham fitting and Metrics
# ===================================
# Now that we have some ODFs, let us fit the Bingham functions to them by using
# the function `sf_to_bingham`:

# A maximum search angle of 45 degrees is chosen arbitrarily for fitting
# each ODF lobe.
max_search_angle = 45
BinghamMetrics = sf_to_bingham(csd_odf, sphere, max_search_angle)

###############################################################################
# The above function outputs a `BinghamMetrics` class instance, containing the
# parameters of the fitted Bingham functions. The metrics of interest contained
# in the `BinghamMetrics` class instance are:
#
# - amplitude_lobe (the maximum value for each lobe. Also known as Bingham's
#       f_0 parameter.)
# - fd_lobe (fiber densitiy: as defined in :footcite:p:`Riffert2014`,
#       one for each peak.)
# - fs_lobe (fiber spread: as defined in :footcite:p:`Riffert2014`,
#       one for each peak.)
# - fd_voxel (voxel fiber density: average of fd across all ODF lobes.)
# - fs_voxel (voxel fiber spread: average of fs across all ODF lobes.)
# - odi1_lobe (orientation dispersion index along Bingham's first dispersion
#       axis, one for each lobe. Defined in :footcite:p:`NetoHenriques2018`
#       and :footcite:p:`Zhang2012`.)
# - odi2_lobe (orientation dispersion index along Bingham's second dispersion
#       axis, one for each lobe.)
# - odi_total_lobe (orientation dispersion index averaged across both Binghams'
#       dispersion axes. Defined in :footcite:p:`Tariq2016`.)
# - odi1_voxel (orientation dispersion index along Bingham's first dispersion
#       axis, averaged across all lobes)
# - odi2_voxel (orientation dispersion index along Bingham's second dispersion
#       axis, averaged across all lobes)
# - odi_total_voxel (orientation dispersion index averaged across both
#       Binghams' axes, averaged across all lobes)
# - peak_dirs (peak directions in Cartesian coordinates given by the Bingham
#       fitting, also known as parameter mu_0. These directions are slightly
#       different than the peak directions given by the function
#       `peaks_from_model`.)
#
# For illustration purposes, the fitted Bingham derived metrics can be
# visualized using the following lines of code:

bim_odf = BinghamMetrics.odf(sphere)

scene.rm(fodf_spheres)

fodf_spheres = actor.odf_slicer(
    bim_odf, sphere=sphere, scale=0.9, norm=False, colormap="plasma"
)
scene.add(fodf_spheres)

print("Saving the illustration as Bingham_odfs.png")
window.record(scene, out_path="Bingham_odfs.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Bingham functions fitted to CSD fiber ODFs.
#
# Alternatively to fitting Bingham functions to sampled ODFs, DIPY also
# contains the function `sh_to_bingham` to perform Bingham fitting from the
# ODF's spherical harmonic representation. Although this process may require
# longer processing times, this function may be useful to avoid memory issues
# in handling heavily sampled ODFs. For example, you may have reconstructed
# ODFs using another script and saved their spherical harmonics to disk.
# This function is for such cases. Below we show the lines of code to use the
# function `sh_to_bingham` (feel free to skip these lines if the function
# `sf_to_bingham` worked fine for you). Note, to use `sh_to_bingham` you
# need to specify the maximum order of spherical harmonics that you defined
# when reconstructing the ODF. In this example this was set to 8 for
# the function `csd_model`:

sh_coeff = csd_fit.shm_coeff
BinghamMetrics = sh_to_bingham(sh_coeff, sphere, max_search_angle)

###############################################################################
# Step 3. Bingham Metrics
# =======================
# As mentioned above, reconstructed Bingham functions can be useful to
# quantify properties from ODFs :footcite:p:`Riffert2014`,
# :footcite:p:`NetoHenriques2018`. Below we plot the Bingham metrics
# expected to be proportional to the fiber density (FD) of specific fiber
# populations.

FD_ODF_l1 = BinghamMetrics.fd_lobe[:, :, 0, 0]
FD_ODF_l2 = BinghamMetrics.fd_lobe[:, :, 0, 1]
FD_voxel = BinghamMetrics.fd_voxel[:, :, 0]

FD_images = [FD_ODF_l1[:, -1:1:-1].T, FD_ODF_l2[:, -1:1:-1].T, FD_voxel[:, -1:1:-1].T]
FD_labels = ["FD ODF lobe 1", "FD ODF lobe 2", "FD ODF voxel"]
kwargs = [{"vmin": 0, "vmax": 2}, {"vmin": 0, "vmax": 2}, {"vmin": 0, "vmax": 2}]

print("Saving the illustration as Bingham_fd.png")
image_mosaic(
    FD_images,
    ax_labels=FD_labels,
    ax_kwargs=kwargs,
    figsize=(16, 4),
    filename="Bingham_fd.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# The figure shows from left to right: 1) the FD estimated for the
# first ODF peak (showing larger values in white matter); 2) the FD estimated
# for the second ODF peak (showing non-zero values in regions of crossing
# white matter fibers); and 3) the sum of FD estimates across all ODF lobes
# (quantity that should be proportional to the density of all fibers within
# each voxel).
#
# Bingham functions can also be used to quantify fiber dispersion from the
# ODFs :footcite:p:`NetoHenriques2018`. In addition to quantifying a combined
# orientation dispersion index (`ODI_total`) for each ODF lobe
# :footcite:p:`Tariq2016`, Bingham functions allow  the quantification of
# dispersion along two main axes (`ODI_1` and `ODI_2`), offering unique
# information of fiber orientation variability within the brain tissue. Below
# we show how to extract these indexes from the largest ODF peak. Note, for
# better visualization of ODI estimates, voxels with total FD lower than 0.5
# are masked.

ODIt = BinghamMetrics.odi_total_lobe[:, :, 0, 0]
ODI1 = BinghamMetrics.odi1_lobe[:, :, 0, 0]
ODI2 = BinghamMetrics.odi2_lobe[:, :, 0, 0]

ODIt[FD_voxel < 0.5] = 0
ODI1[FD_voxel < 0.5] = 0
ODI2[FD_voxel < 0.5] = 0

ODI_images = [ODI1[:, -1:1:-1].T, ODI2[:, -1:1:-1].T, ODIt[:, -1:1:-1].T]
ODI_labels = ["ODI_1 (lobe 1)", "ODI_2 (lobe 1)", "ODI_total (lobe 1)"]
kwargs = [{"vmin": 0, "vmax": 0.2}, {"vmin": 0, "vmax": 0.2}, {"vmin": 0, "vmax": 0.2}]
print("Saving the illustration as Bingham_ODI_lobe1.png")
image_mosaic(
    ODI_images,
    ax_labels=ODI_labels,
    ax_kwargs=kwargs,
    figsize=(15, 5),
    filename="Bingham_ODI_lobe1.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# The figure shows from left to right: 1) ODI of the largest ODF lobe along
# the axis with greater dispersion, a.k.a. ODI_1 (direction in which fibers
# exhibit the most variability in orientation); 2) ODI of the largest ODF lobe
# along the axis with lesser dispersion, a.k.a ODI_2 (directions in which
# fiber orientations are more uniform); and 3) total ODI of the largest lobe
# across both axes.
#
# Above, we focused on the largest ODF's lobe, representing the most pronounced
# fiber population within a voxel. However, this methodology is not limited to
# a singular lobe since it can be applied to the other ODF lobes. Below, we
# show the analogous figures for the second-largest ODF lobe. Note that for
# this figure, regions of white matter that contain only a single fiber
# population display ODI estimates of zero, corresponding to ODF profiles
# lacking a second ODF lobe.

ODIt = BinghamMetrics.odi_total_lobe[:, :, 0, 1]
ODI1 = BinghamMetrics.odi1_lobe[:, :, 0, 1]
ODI2 = BinghamMetrics.odi2_lobe[:, :, 0, 1]

ODIt[FD_voxel < 0.5] = 0
ODI1[FD_voxel < 0.5] = 0
ODI2[FD_voxel < 0.5] = 0

ODI_images = [ODI1[:, -1:1:-1].T, ODI2[:, -1:1:-1].T, ODIt[:, -1:1:-1].T]
ODI_labels = ["ODI_1 (lobe 2)", "ODI_2 (lobe 2)", "ODI_voxel (lobe 2)"]
kwargs = [{"vmin": 0, "vmax": 0.2}, {"vmin": 0, "vmax": 0.2}, {"vmin": 0, "vmax": 0.2}]
print("Saving the illustration as Bingham_ODI_lobe2.png")
image_mosaic(
    ODI_images,
    ax_labels=ODI_labels,
    ax_kwargs=kwargs,
    figsize=(15, 5),
    filename="Bingham_ODI_lobe2.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# The figure shows from left to right: 1) ODI for the second-largest ODF lobe
# along the axis with greater dispersion a.k.a. ODI_1 (direction in which
# fibers exhibit the most variability in orientation); 2) ODI for the
# second-largest ODF lobe along the axis with lesser dispersion a.k.a. ODI_2
# (directions in which fiber orientations are more uniform); and 3) total ODI
# for the second-largest ODF lobe across both axes. In this figure, regions of
# the white matter that contain only a single fiber population (one ODF lobe)
# display ODI estimates of zero, corresponding to ODF profiles lacking a
# second ODF lobe.
#
# BinghamMetrics can also be used to compute the average ODI quantities across
# all ODF lobes a.k.a. voxel ODI (see below). The average quantitaties are
# computed by weigthing each ODF lobe with their respective fiber density (FD)
# value. These quantities are plotted in the following figure.

ODIt = BinghamMetrics.odi_total_voxel[:, :, 0]
ODI1 = BinghamMetrics.odi1_voxel[:, :, 0]
ODI2 = BinghamMetrics.odi2_voxel[:, :, 0]

ODIt[FD_voxel < 0.5] = 0
ODI1[FD_voxel < 0.5] = 0
ODI2[FD_voxel < 0.5] = 0

ODI_images = [ODI1[:, -1:1:-1].T, ODI2[:, -1:1:-1].T, ODIt[:, -1:1:-1].T]
ODI_labels = ["ODI_1 (voxel)", "ODI_2 (voxel)", "ODI_total (voxel)"]
kwargs = [{"vmin": 0, "vmax": 0.2}, {"vmin": 0, "vmax": 0.2}, {"vmin": 0, "vmax": 0.2}]
print("Saving the illustration as Bingham_ODI.png")
image_mosaic(
    ODI_images,
    ax_labels=ODI_labels,
    ax_kwargs=kwargs,
    figsize=(15, 5),
    filename="Bingham_ODI_voxel.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# The figure shows from left to right: 1) weighted-averaged ODI_1 across all ODF
# lobes; 2) weighted-averaged ODI_2 across all ODF lobe; 3) weighted-averaged
# ODI_total across all ODF lobes.
#
# References
# ----------
#
# .. footbibliography::
#



import logging

import numpy as np

from dipy.segment.clustering import qbx_and_merge
from dipy.testing.decorators import warning_for_keywords
from dipy.tracking.streamline import length
from dipy.utils.optpkg import optional_package

fury, has_fury, setup_module = optional_package("fury", min_version="0.10.0")

if has_fury:
    from fury import actor
    from fury.lib import VTK_OBJECT, calldata_type
    from fury.shaders import add_shader_callback, shader_to_actor


class ClustersVisualizer:
    @warning_for_keywords()
    def __init__(self, show_manager, scene, tractograms, *, enable_callbacks=True):
        # TODO: Avoid passing the entire show manager to the visualizer
        self.__show_man = show_manager
        self.__scene = scene
        self.__tractograms = tractograms
        self.__enable_callbacks = enable_callbacks

        self.__tractogram_clusters = {}

        self.__first_time = True
        self.__tractogram_colors = []

        self.__centroid_actors = {}
        self.__cluster_actors = {}

        self.__lengths = []
        self.__sizes = []

    def __apply_shader(self, dict_element):
        decl = """
            uniform float selected;
            """

        impl = """
            if (selected == 1)
            {
                fragOutput0 += vec4(.2, .2, .2, 0);
            }
            """
        shader_to_actor(dict_element["actor"], "fragment", decl_code=decl)
        shader_to_actor(
            dict_element["actor"], "fragment", impl_code=impl, block="light"
        )

        @calldata_type(VTK_OBJECT)
        def uniform_selected_callback(caller, event, calldata=None):
            program = calldata
            if program is not None:
                program.SetUniformf("selected", dict_element["selected"])

        add_shader_callback(
            dict_element["actor"], uniform_selected_callback, priority=100
        )

    def __left_click_centroid_callback(self, obj, event):
        self.__centroid_actors[obj]["selected"] = not self.__centroid_actors[obj][
            "selected"
        ]
        self.__cluster_actors[self.__centroid_actors[obj]["actor"]]["selected"] = (
            self.__centroid_actors[obj]["selected"]
        )
        # TODO: Find another way to rerender
        self.__show_man.render()

    def __left_click_cluster_callback(self, obj, event):
        if self.__cluster_actors[obj]["selected"]:
            self.__cluster_actors[obj]["actor"].VisibilityOn()
            ca = self.__cluster_actors[obj]["actor"]
            self.__centroid_actors[ca]["selected"] = 0
            obj.VisibilityOff()
            self.__centroid_actors[ca]["expanded"] = 0
        # TODO: Find another way to rerender
        self.__show_man.render()

    def add_cluster_actors(self, tract_idx, streamlines, thr, colors):
        # Saving the tractogram colors in case of reclustering
        if self.__first_time:
            self.__tractogram_colors.append(colors)

        logging.info(f"Clustering threshold {thr}")
        clusters = qbx_and_merge(streamlines, [40, 30, 25, 20, thr])
        self.__tractogram_clusters[tract_idx] = clusters
        centroids = clusters.centroids
        logging.info(f"Total number of centroids = {len(centroids)}")

        lengths = [length(c) for c in centroids]
        self.__lengths.extend(lengths)
        lengths = np.array(lengths)

        sizes = [len(c) for c in clusters]
        self.__sizes.extend(sizes)
        sizes = np.array(sizes)
        linewidths = np.interp(sizes, [np.min(sizes), np.max(sizes)], [0.1, 2.0])

        logging.info(f"Minimum number of streamlines in cluster {np.min(sizes)}")
        logging.info(f"Maximum number of streamlines in cluster {np.max(sizes)}")

        logging.info("Building cluster actors\n")
        for idx, cent in enumerate(centroids):
            centroid_actor = actor.streamtube(
                [cent], colors=colors, linewidth=linewidths[idx], lod=False
            )
            self.__scene.add(centroid_actor)

            cluster_actor = actor.line(clusters[idx][:], lod=False)
            cluster_actor.GetProperty().SetRenderLinesAsTubes(1)
            cluster_actor.GetProperty().SetLineWidth(6)
            cluster_actor.GetProperty().SetOpacity(1)
            cluster_actor.VisibilityOff()
            self.__scene.add(cluster_actor)

            # Every centroid actor is paired to a cluster actor
            self.__centroid_actors[centroid_actor] = {
                "actor": cluster_actor,
                "cluster": idx,
                "tractogram": tract_idx,
                "size": sizes[idx],
                "length": lengths[idx],
                "selected": 0,
                "expanded": 0,
            }

            self.__cluster_actors[cluster_actor] = {
                "actor": centroid_actor,
                "cluster": idx,
                "tractogram": tract_idx,
                "size": sizes[idx],
                "length": lengths[idx],
                "selected": 0,
                "highlighted": 0,
            }

            self.__apply_shader(self.__centroid_actors[centroid_actor])
            self.__apply_shader(self.__cluster_actors[cluster_actor])

            if self.__enable_callbacks:
                centroid_actor.AddObserver(
                    "LeftButtonPressEvent", self.__left_click_centroid_callback, 1.0
                )
                cluster_actor.AddObserver(
                    "LeftButtonPressEvent", self.__left_click_cluster_callback, 1.0
                )

    def recluster_tractograms(self, thr):
        for cent in self.__centroid_actors:
            self.__scene.rm(self.__centroid_actors[cent]["actor"])
        for clus in self.__cluster_actors:
            self.__scene.rm(self.__cluster_actors[clus]["actor"])
        self.__tractogram_clusters = {}
        self.__centroid_actors = {}
        self.__cluster_actors = {}
        self.__lengths = []
        self.__sizes = []

        # Keeping states of some attributes
        self.__first_time = False

        for t, sft in enumerate(self.__tractograms):
            streamlines = sft.streamlines
            self.add_cluster_actors(t, streamlines, thr, self.__tractogram_colors[t])

    @property
    def centroid_actors(self):
        return self.__centroid_actors

    @property
    def cluster_actors(self):
        return self.__cluster_actors

    @property
    def lengths(self):
        return np.array(self.__lengths)

    @property
    def sizes(self):
        return np.array(self.__sizes)

    @property
    def tractogram_clusters(self):
        return self.__tractogram_clusters



import logging
import os
from os.path import join as pjoin
from tempfile import TemporaryDirectory
import warnings

import numpy as np
import numpy.testing as npt
import pytest

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, load_nifti_data, save_nifti
from dipy.io.peaks import load_pam
from dipy.reconst.shm import descoteaux07_legacy_msg, sph_harm_ind_list
from dipy.sims.voxel import multi_tensor
from dipy.utils.optpkg import optional_package
from dipy.workflows.reconst import (
    ReconstForecastFlow,
    ReconstGQIFlow,
    ReconstRUMBAFlow,
    ReconstSFMFlow,
)

_, has_sklearn, _ = optional_package("sklearn")
logging.getLogger().setLevel(logging.INFO)


def simulate_multitensor_data(gtab):
    dwi = np.zeros((2, 2, 2, len(gtab.bvals)))
    # Diffusion of tissue and water compartments are constant for all voxel
    mevals = np.array([[0.0017, 0.0003, 0.0003], [0.003, 0.003, 0.003]])
    # volume fractions
    GTF = np.array([[[0.06, 0.71], [0.33, 0.91]], [[0.0, 0.0], [0.0, 0.0]]])
    for i in range(2):
        for j in range(2):
            gtf = GTF[0, i, j]
            S, p = multi_tensor(
                gtab,
                mevals,
                S0=100,
                angles=[(90, 0), (90, 0)],
                fractions=[(1 - gtf) * 100, gtf * 100],
                snr=None,
            )
            dwi[0, i, j] = S
    return dwi


def test_reconst_rumba():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        reconst_flow_core(ReconstRUMBAFlow)


@pytest.mark.skipif(not has_sklearn, reason="Requires sklearn")
def test_reconst_sfm():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        reconst_flow_core(ReconstSFMFlow)


def test_reconst_gqi():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        reconst_flow_core(ReconstGQIFlow)


def test_reconst_forecast():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        reconst_flow_core(ReconstForecastFlow, use_multishell_data=True)


def reconst_flow_core(flow, *, use_multishell_data=None, **kwargs):
    with TemporaryDirectory() as out_dir:
        data_path, bval_path, bvec_path = get_fnames(name="small_64D")
        if use_multishell_data:
            bvals, bvecs = read_bvals_bvecs(bval_path, bvec_path)

            # FW model requires multishell data
            bvals_2s = np.concatenate((bvals, bvals * 1.5), axis=0)
            bvecs_2s = np.concatenate((bvecs, bvecs), axis=0)
            gtab_2s = gradient_table(bvals_2s, bvecs=bvecs_2s)
            bval_path = pjoin(out_dir, os.path.basename(bval_path))
            bvec_path = pjoin(out_dir, os.path.basename(bvec_path))
            np.savetxt(bval_path, bvals_2s)
            np.savetxt(bvec_path, bvecs_2s)

            volume = simulate_multitensor_data(gtab_2s)
            data_path = pjoin(out_dir, "dwi.nii.gz")
            mask = np.ones_like(volume[..., 0], dtype=bool)
            mask_path = pjoin(out_dir, "mask.nii.gz")
            save_nifti(mask_path, mask.astype(np.int32), np.eye(4))
            save_nifti(data_path, volume, np.eye(4))
        else:
            volume, affine = load_nifti(data_path)
            mask = np.ones_like(volume[:, :, :, 0])
            mask_path = pjoin(out_dir, "tmp_mask.nii.gz")
            save_nifti(mask_path, mask.astype(np.uint8), affine)

        reconst_flow = flow()
        for sh_order_max in [
            8,
        ]:
            reconst_flow.run(
                data_path,
                bval_path,
                bvec_path,
                mask_path,
                sh_order_max=sh_order_max,
                out_dir=out_dir,
                extract_pam_values=True,
                **kwargs,
            )

            gfa_path = reconst_flow.last_generated_outputs["out_gfa"]
            gfa_data = load_nifti_data(gfa_path)
            npt.assert_equal(gfa_data.shape, volume.shape[:-1])

            peaks_dir_path = reconst_flow.last_generated_outputs["out_peaks_dir"]
            peaks_dir_data = load_nifti_data(peaks_dir_path)
            npt.assert_equal(peaks_dir_data.shape[-1], 15)
            npt.assert_equal(peaks_dir_data.shape[:-1], volume.shape[:-1])

            peaks_idx_path = reconst_flow.last_generated_outputs["out_peaks_indices"]
            peaks_idx_data = load_nifti_data(peaks_idx_path)
            npt.assert_equal(peaks_idx_data.shape[-1], 5)
            npt.assert_equal(peaks_idx_data.shape[:-1], volume.shape[:-1])

            peaks_vals_path = reconst_flow.last_generated_outputs["out_peaks_values"]
            peaks_vals_data = load_nifti_data(peaks_vals_path)
            npt.assert_equal(peaks_vals_data.shape[-1], 5)
            npt.assert_equal(peaks_vals_data.shape[:-1], volume.shape[:-1])

            shm_path = reconst_flow.last_generated_outputs["out_shm"]
            shm_data = load_nifti_data(shm_path)
            # Test that the number of coefficients is what you would expect
            # given the order of the sh basis:
            npt.assert_equal(
                shm_data.shape[-1], sph_harm_ind_list(sh_order_max)[0].shape[0]
            )
            npt.assert_equal(shm_data.shape[:-1], volume.shape[:-1])

            pam = load_pam(reconst_flow.last_generated_outputs["out_pam"])
            npt.assert_allclose(
                pam.peak_dirs.reshape(peaks_dir_data.shape), peaks_dir_data
            )

            npt.assert_allclose(pam.peak_values, peaks_vals_data)
            npt.assert_allclose(pam.peak_indices, peaks_idx_data)
            npt.assert_allclose(pam.shm_coeff, shm_data, atol=1e-7)
            npt.assert_allclose(pam.gfa, gfa_data)



"""Testing convert utilities."""

import numpy as np

from dipy.utils.convert import expand_range


def test_expand_range():
    test_cases = [
        ("1,2,3,4", np.array([1, 2, 3, 4])),
        ("5-7", np.array([5, 6, 7])),
        ("0", np.array([0])),
        ("0,", np.array([0])),
        ("0, ", np.array([0])),
        ("1,2,5-7,10", np.array([1, 2, 5, 6, 7, 10])),
        ("1,2,5-7,10", [1, 2, 5, 6, 7, 10]),
        ("1,2,5-7,10", (1, 2, 5, 6, 7, 10)),
        ("1,a,3", ValueError),
        ("1-2-3", ValueError),
    ]

    for input_string, expected_output in test_cases:
        # print(f"Running test for input: {input_string}")
        if not isinstance(expected_output, (np.ndarray, list, tuple)):
            try:
                expand_range(input_string)
                raise AssertionError(f"Expected ValueError for input: {input_string}")
            except ValueError:
                assert True
        else:
            result = expand_range(input_string)
            np.testing.assert_array_equal(result, expected_output)
            # print(f"Test passed for input: {input_string}")



# cython: wraparound=False, cdivision=True, boundscheck=False

import numpy as np
cimport numpy as cnp

from libc.math cimport sqrt, acos

from dipy.segment.cythonutils cimport tuple2shape, shape2tuple, same_shape
from dipy.segment.featurespeed cimport IdentityFeature

DEF biggest_double = 1.7976931348623157e+308  #  np.finfo('f8').max

import math
cdef double PI = math.pi


cdef class Metric:
    """ Computes a distance between two sequential data.

    A sequence of N-dimensional points is represented as a 2D array with
    shape (nb_points, nb_dimensions). A `feature` object can be specified
    in order to calculate the distance between extracted features, rather
    than directly between the sequential data.

    Parameters
    ----------
    feature : `Feature` object, optional
        It is used to extract features before computing the distance.

    Notes
    -----
    When subclassing `Metric`, one only needs to override the `dist` and
    `are_compatible` methods.
    """
    def __init__(Metric self, Feature feature=IdentityFeature()):
        self.feature = feature
        self.is_order_invariant = self.feature.is_order_invariant

    property feature:
        """ `Feature` object used to extract features from sequential data """
        def __get__(Metric self):
            return self.feature

    property is_order_invariant:
        """ Is this metric invariant to the sequence's ordering """
        def __get__(Metric self):
            return bool(self.is_order_invariant)

    cdef int c_are_compatible(Metric self, Shape shape1, Shape shape2) except -1 nogil:
        """ Cython version of `Metric.are_compatible`. """
        with gil:
            return self.are_compatible(shape2tuple(shape1), shape2tuple(shape2))

    cdef double c_dist(Metric self, Data2D features1, Data2D features2) except -1 nogil:
        """ Cython version of `Metric.dist`. """
        with gil:
            _features1 = np.asarray(<float[:features1.shape[0], :features1.shape[1]]> <float*> features1._data)
            _features2 = np.asarray(<float[:features2.shape[0], :features2.shape[1]]> <float*> features2._data)
            return self.dist(_features1, _features2)

    cpdef are_compatible(Metric self, shape1, shape2):
        """ Checks if features can be used by `metric.dist` based on their shape.

        Basically this method exists so we don't have to do this check
        inside the `metric.dist` function (speedup).

        Parameters
        ----------
        shape1 : int, 1-tuple or 2-tuple
            shape of the first data point's features
        shape2 : int, 1-tuple or 2-tuple
            shape of the second data point's features

        Returns
        -------
        are_compatible : bool
            whether or not shapes are compatible
        """
        raise NotImplementedError("Metric's subclasses must implement method `are_compatible(self, shape1, shape2)`!")

    cpdef double dist(Metric self, features1, features2) except -1:
        """ Computes a distance between two data points based on their features.

        Parameters
        ----------
        features1 : 2D array
            Features of the first data point.
        features2 : 2D array
            Features of the second data point.

        Returns
        -------
        double
            Distance between two data points.
        """
        raise NotImplementedError("Metric's subclasses must implement method `dist(self, features1, features2)`!")


cdef class CythonMetric(Metric):
    """ Computes a distance between two sequential data.

    A sequence of N-dimensional points is represented as a 2D array with
    shape (nb_points, nb_dimensions). A `feature` object can be specified
    in order to calculate the distance between extracted features, rather
    than directly between the sequential data.

    Parameters
    ----------
    feature : `Feature` object, optional
        It is used to extract features before computing the distance.

    Notes
    -----
    When subclassing `CythonMetric`, one only needs to override the `c_dist` and
    `c_are_compatible` methods.
    """
    cpdef are_compatible(CythonMetric self, shape1, shape2):
        """ Checks if features can be used by `metric.dist` based on their shape.

        Basically this method exists so we don't have to do this check
        inside method `dist` (speedup).

        Parameters
        ----------
        shape1 : int, 1-tuple or 2-tuple
            Shape of the first data point's features.
        shape2 : int, 1-tuple or 2-tuple
            Shape of the second data point's features.

        Returns
        -------
        bool
            Whether or not shapes are compatible.

        Notes
        -----
        This method calls its Cython version `self.c_are_compatible` accordingly.
        """
        if np.asarray(shape1).ndim == 0:
            shape1 = (1, shape1)
        elif len(shape1) == 1:
            shape1 = (1,) + shape1

        if np.asarray(shape2).ndim == 0:
            shape2 = (1, shape2)
        elif len(shape2) == 1:
            shape2 = (1,) + shape2

        return self.c_are_compatible(tuple2shape(shape1), tuple2shape(shape2)) == 1

    cpdef double dist(CythonMetric self, features1, features2) except -1:
        """ Computes a distance between two data points based on their features.

        Parameters
        ----------
        features1 : 2D array
            Features of the first data point.
        features2 : 2D array
            Features of the second data point.

        Returns
        -------
        double
            Distance between two data points.

        Notes
        -----
        This method calls its Cython version `self.c_dist` accordingly.
        """
        # If needed, we convert features to 2D arrays.
        features1 = np.asarray(features1)
        if features1.ndim == 0:
            features1 = features1[np.newaxis, np.newaxis]
        elif features1.ndim == 1:
            features1 = features1[np.newaxis]
        elif features1.ndim == 2:
            pass
        else:
            raise TypeError("Only scalar, 1D or 2D array features are"
                            " supported for parameter 'features1'!")

        features2 = np.asarray(features2)
        if features2.ndim == 0:
            features2 = features2[np.newaxis, np.newaxis]
        elif features2.ndim == 1:
            features2 = features2[np.newaxis]
        elif features2.ndim == 2:
            pass
        else:
            raise TypeError("Only scalar, 1D or 2D array features are"
                            " supported for parameter 'features2'!")

        if not self.are_compatible(features1.shape, features2.shape):
            raise ValueError("Features are not compatible according to this metric!")

        return self.c_dist(features1, features2)


cdef class SumPointwiseEuclideanMetric(CythonMetric):
    r""" Computes the sum of pointwise Euclidean distances between two
    sequential data.

    A sequence of N-dimensional points is represented as a 2D array with
    shape (nb_points, nb_dimensions). A `feature` object can be specified
    in order to calculate the distance between the features, rather than
    directly between the sequential data.

    Parameters
    ----------
    feature : `Feature` object, optional
        It is used to extract features before computing the distance.

    Notes
    -----
    The distance between two 2D sequential data::

        s1       s2

        0*   a    *0
          \       |
           \      |
           1*     |
            |  b  *1
            |      \
            2*      \
                c    *2

    is equal to $a+b+c$ where $a$ is the Euclidean distance between s1[0] and
    s2[0], $b$ between s1[1] and s2[1] and $c$ between s1[2] and s2[2].
    """
    cdef double c_dist(SumPointwiseEuclideanMetric self, Data2D features1, Data2D features2) except -1 nogil:
        cdef :
            int N = features1.shape[0], D = features1.shape[1]
            int n, d
            double dd, dist_n, dist = 0.0

        for n in range(N):
            dist_n = 0.0
            for d in range(D):
                dd = features1[n, d] - features2[n, d]
                dist_n += dd*dd

            dist += sqrt(dist_n)

        return dist

    cdef int c_are_compatible(SumPointwiseEuclideanMetric self, Shape shape1, Shape shape2) except -1 nogil:
        return same_shape(shape1, shape2)


cdef class AveragePointwiseEuclideanMetric(SumPointwiseEuclideanMetric):
    r""" Computes the average of pointwise Euclidean distances between two
    sequential data.

    A sequence of N-dimensional points is represented as a 2D array with
    shape (nb_points, nb_dimensions). A `feature` object can be specified
    in order to calculate the distance between the features, rather than
    directly between the sequential data.

    Parameters
    ----------
    feature : `Feature` object, optional
        It is used to extract features before computing the distance.

    Notes
    -----
    The distance between two 2D sequential data::

        s1       s2

        0*   a    *0
          \       |
           \      |
           1*     |
            |  b  *1
            |      \
            2*      \
                c    *2

    is equal to $(a+b+c)/3$ where $a$ is the Euclidean distance between s1[0] and
    s2[0], $b$ between s1[1] and s2[1] and $c$ between s1[2] and s2[2].
    """
    cdef double c_dist(AveragePointwiseEuclideanMetric self, Data2D features1, Data2D features2) except -1 nogil:
        cdef int N = features1.shape[0]
        cdef double dist = SumPointwiseEuclideanMetric.c_dist(self, features1, features2)
        return dist / N


cdef class MinimumAverageDirectFlipMetric(AveragePointwiseEuclideanMetric):
    r""" Computes the MDF distance (minimum average direct-flip) between two
    sequential data.

    A sequence of N-dimensional points is represented as a 2D array with
    shape (nb_points, nb_dimensions).

    Notes
    -----
    The distance between two 2D sequential data::

        s1       s2

        0*   a    *0
          \       |
           \      |
           1*     |
            |  b  *1
            |      \
            2*      \
                c    *2

    is equal to $\min((a+b+c)/3, (a'+b'+c')/3)$ where $a$ is the Euclidean distance
    between s1[0] and s2[0], $b$ between s1[1] and s2[1], $c$ between s1[2]
    and s2[2], $a'$ between s1[0] and s2[2], $b'$ between s1[1] and s2[1]
    and $c'$ between s1[2] and s2[0].
    """
    property is_order_invariant:
        """ Is this metric invariant to the sequence's ordering """
        def __get__(MinimumAverageDirectFlipMetric self):
            return True  # Ordering is handled in the distance computation

    cdef double c_dist(MinimumAverageDirectFlipMetric self, Data2D features1, Data2D features2) except -1 nogil:
        cdef double dist_direct = AveragePointwiseEuclideanMetric.c_dist(self, features1, features2)
        cdef double dist_flipped = AveragePointwiseEuclideanMetric.c_dist(self, features1, features2[::-1])
        return min(dist_direct, dist_flipped)


cdef class CosineMetric(CythonMetric):
    r""" Computes the cosine distance between two vectors.

    A vector (i.e. a N-dimensional point) is represented as a 2D array with
    shape (1, nb_dimensions).

    Notes
    -----
    The distance between two vectors $v_1$ and $v_2$ is equal to
    $\frac{1}{\pi} \arccos\left(\frac{v_1 \cdot v_2}{\|v_1\| \|v_2\|}\right)$
    and is bounded within $[0,1]$.
    """
    def __init__(CosineMetric self, Feature feature):
        super(CosineMetric, self).__init__(feature=feature)

    cdef int c_are_compatible(CosineMetric self, Shape shape1, Shape shape2) except -1 nogil:
        return same_shape(shape1, shape2) != 0 and shape1.dims[0] == 1

    cdef double c_dist(CosineMetric self, Data2D features1, Data2D features2) except -1 nogil:
        cdef :
            int d, D = features1.shape[1]
            double sqr_norm_features1 = 0.0, sqr_norm_features2 = 0.0
            double cos_theta = 0.0

        for d in range(D):
            cos_theta += features1[0, d] * features2[0, d]
            sqr_norm_features1 += features1[0, d] * features1[0, d]
            sqr_norm_features2 += features2[0, d] * features2[0, d]

        if sqr_norm_features1 == 0.:
            if sqr_norm_features2 == 0.:
                return 0.
            else:
                return 1.

        cos_theta /= sqrt(sqr_norm_features1) * sqrt(sqr_norm_features2)

        # Make sure it's in [-1, 1], i.e. within domain of arccosine
        cos_theta = min(cos_theta, 1.)
        cos_theta = max(cos_theta, -1.)
        return acos(cos_theta) / PI  # Normalized cosine distance


cpdef distance_matrix(Metric metric, data1, data2=None):
    """ Computes the distance matrix between two lists of sequential data.

    The distance matrix is obtained by computing the pairwise distance of all
    tuples spawn by the Cartesian product of `data1` with `data2`. If `data2`
    is not provided, the Cartesian product of `data1` with itself is used
    instead. A sequence of N-dimensional points is represented as a 2D array with
    shape (nb_points, nb_dimensions).

    Parameters
    ----------
    metric : `Metric` object
        Tells how to compute the distance between two sequential data.
    data1 : list of 2D arrays
        List of sequences of N-dimensional points.
    data2 : list of 2D arrays
        Llist of sequences of N-dimensional points.

    Returns
    -------
    2D array (double)
        Distance matrix.
    """
    cdef cnp.npy_intp i, j
    if data2 is None:
        data2 = data1

    shape = metric.feature.infer_shape(data1[0].astype(np.float32))
    distance_matrix = np.zeros((len(data1), len(data2)), dtype=np.float64)
    cdef:
        Data2D features1 = np.empty(shape, np.float32)
        Data2D features2 = np.empty(shape, np.float32)

    for i in range(len(data1)):
        datum1 = data1[i] if data1[i].flags.writeable and data1[i].dtype is np.float32 else data1[i].astype(np.float32)
        metric.feature.c_extract(datum1, features1)
        for j in range(len(data2)):
            datum2 = data2[j] if data2[j].flags.writeable and data2[j].dtype is np.float32 else data2[j].astype(np.float32)
            metric.feature.c_extract(datum2, features2)
            distance_matrix[i, j] = metric.c_dist(features1, features2)

    return distance_matrix


cpdef double dist(Metric metric, datum1, datum2) except -1:
    """ Computes a distance between `datum1` and `datum2`.

    A sequence of N-dimensional points is represented as a 2D array with
    shape (nb_points, nb_dimensions).

    Parameters
    ----------
    metric : `Metric` object
        Tells how to compute the distance between `datum1` and `datum2`.
    datum1 : 2D array
        Sequence of N-dimensional points.
    datum2 : 2D array
        Sequence of N-dimensional points.

    Returns
    -------
    double
        Distance between two data points.
    """
    datum1 = datum1 if datum1.flags.writeable and datum1.dtype is np.float32 else datum1.astype(np.float32)
    datum2 = datum2 if datum2.flags.writeable and datum2.dtype is np.float32 else datum2.astype(np.float32)

    cdef:
        Shape shape1 = metric.feature.c_infer_shape(datum1)
        Shape shape2 = metric.feature.c_infer_shape(datum2)
        Data2D features1 = np.empty(shape2tuple(shape1), np.float32)
        Data2D features2 = np.empty(shape2tuple(shape2), np.float32)

    metric.feature.c_extract(datum1, features1)
    metric.feature.c_extract(datum2, features2)
    return metric.c_dist(features1, features2)



import logging
from warnings import warn

import numpy as np
from packaging.version import Version

from dipy import __version__ as horizon_version
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.streamline import save_tractogram
from dipy.testing.decorators import warning_for_keywords
from dipy.tracking.streamline import Streamlines
from dipy.utils.optpkg import optional_package
from dipy.viz.gmem import GlobalHorizon
from dipy.viz.horizon.tab import (
    ClustersTab,
    PeaksTab,
    ROIsTab,
    SlicesTab,
    SurfaceTab,
    TabManager,
    build_label,
)
from dipy.viz.horizon.util import (
    check_img_dtype,
    check_img_shapes,
    check_peak_size,
    unpack_image,
    unpack_surface,
)
from dipy.viz.horizon.visualizer import (
    ClustersVisualizer,
    PeaksVisualizer,
    SlicesVisualizer,
    SurfaceVisualizer,
)

fury, has_fury, setup_module = optional_package("fury", min_version="0.10.0")

if has_fury:
    from fury import __version__ as fury_version, actor, ui, window
    from fury.colormap import distinguishable_colormap


# TODO: Re-enable >> right click: see menu
HELP_MESSAGE = """
>> left click: select centroid
>> e: expand centroids
>> r: collapse all clusters
>> h: hide unselected centroids
>> i: invert selection
>> a: select all centroids
>> s: save in file
>> y: new window
>> o: hide/show this panel
"""

SHORTCUT_MESSAGE = """
>> shift + drag: move
>> ctrl/cmd + drag: rotate
>> ctrl/cmd + r: reset
>> o: hide/show this panel
"""


class Horizon:
    @warning_for_keywords()
    def __init__(
        self,
        *,
        tractograms=None,
        images=None,
        pams=None,
        surfaces=None,
        cluster=False,
        rgb=False,
        cluster_thr=15.0,
        random_colors=None,
        length_gt=0,
        length_lt=1000,
        clusters_gt=0,
        clusters_lt=10000,
        world_coords=True,
        interactive=True,
        out_png="tmp.png",
        recorded_events=None,
        return_showm=False,
        bg_color=(0, 0, 0),
        order_transparent=True,
        buan=False,
        buan_colors=None,
        roi_images=False,
        roi_colors=(1, 0, 0),
        surface_colors=((1, 0, 0),),
    ):
        """Interactive medical visualization - Invert the Horizon!
        :footcite:p:`Garyfallidis2019`.

        Parameters
        ----------
        tractograms : sequence of StatefulTractograms
            StatefulTractograms are used for making sure that the coordinate
            systems are correct
        images : sequence of tuples
            Each tuple contains data and affine
        pams : sequence of PeakAndMetrics
            Contains peak directions and spherical harmonic coefficients
        surfaces : sequence of tuples
            Each tuple contains vertices and faces
        cluster : bool
            Enable QuickBundlesX clustering
        rgb : bool, optional
            Enable the color image (rgb only, alpha channel will be ignored).
        cluster_thr : float
            Distance threshold used for clustering. Default value 15.0 for
            small animal data you may need to use something smaller such
            as 2.0. The threshold is in mm. For this parameter to be active
            ``cluster`` should be enabled.
        random_colors : string, optional
            Given multiple tractograms and/or ROIs then each tractogram and/or
            ROI will be shown with a different color. If no value is provided,
            both the tractograms and the ROIs will have a different random
            color generated from a distinguishable colormap. If the effect
            should only be applied to one of the 2 types, then use the
            options 'tracts' and 'rois' for the tractograms and the ROIs
            respectively.
        length_gt : float
            Clusters with average length greater than ``length_gt`` amount
            in mm will be shown.
        length_lt : float
            Clusters with average length less than ``length_lt`` amount in mm
            will be shown.
        clusters_gt : int
            Clusters with size greater than ``clusters_gt`` will be shown.
        clusters_lt : int
            Clusters with size less than ``clusters_lt`` will be shown.
        world_coords : bool
            Show data in their world coordinates (not native voxel coordinates)
            Default True.
        interactive : bool
            Allow user interaction. If False then Horizon goes on stealth mode
            and just saves pictures.
        out_png : string
            Filename of saved picture.
        recorded_events : string
            File path to replay recorded events
        return_showm : bool
            Return ShowManager object. Used only at Python level. Can be used
            for extending Horizon's capabilities externally and for testing
            purposes.
        bg_color : ndarray or list or tuple
            Define the background color of the scene.
            Default is black (0, 0, 0)
        order_transparent : bool
            Default True. Use depth peeling to sort transparent objects.
            If True also enables anti-aliasing.
        buan : bool, optional
            Enables BUAN framework visualization. Default is False.
        buan_colors : list, optional
            List of colors for bundles.
        roi_images : bool, optional
            Displays binary images as contours. Default is False.
        roi_colors : ndarray or list or tuple, optional
            Define the colors of the roi images. Default is red (1, 0, 0)


        References
        ----------
        .. footbibliography::
        """
        if not has_fury:
            raise ImportError(
                "Horizon requires FURY. Please install it with pip install fury"
            )
        if Version(fury_version) < Version("0.10.0"):
            ValueError(
                "Horizon requires FURY version 0.10.0 or higher."
                " Please upgrade FURY with pip install -U fury."
            )

        self.cluster = cluster
        self.rgb = rgb
        self.cluster_thr = cluster_thr
        self.random_colors = random_colors
        self.length_lt = length_lt
        self.length_gt = length_gt
        self.clusters_lt = clusters_lt
        self.clusters_gt = clusters_gt
        self.world_coords = world_coords
        self.interactive = interactive
        self.prng = np.random.default_rng(27)
        self.tractograms = tractograms or []
        self.out_png = out_png
        self.images = images or []
        self.pams = pams or []
        self._surfaces = surfaces or []

        self.cea = {}  # holds centroid actors
        self.cla = {}  # holds cluster actors
        self.recorded_events = recorded_events
        self.show_m = None
        self._scene = None
        self.return_showm = return_showm
        self.bg_color = bg_color
        self.order_transparent = order_transparent
        self.buan = buan
        self.buan_colors = buan_colors
        self.__roi_images = roi_images
        self.__roi_colors = roi_colors
        self._surface_colors = surface_colors

        self.color_gen = distinguishable_colormap()

        if self.random_colors is not None:
            if not self.random_colors:
                self.random_colors = ["tracts", "rois"]
        else:
            self.random_colors = []

        self.__clusters_visualizer = None
        self.__tabs = []
        self.__tab_mgr = None

        self.__help_visible = True
        self._shortcut_help_visible = True

        # TODO: Move to another class/module
        self.__hide_centroids = True
        self.__select_all = False

        self._win_size = (0, 0)

    # TODO: Move to another class/module
    def __expand(self):
        centroid_actors = self.__clusters_visualizer.centroid_actors
        lengths = self.__clusters_visualizer.lengths
        sizes = self.__clusters_visualizer.sizes
        min_length = np.min(lengths)
        min_size = np.min(sizes)
        for cent in centroid_actors:
            if centroid_actors[cent]["selected"]:
                if not centroid_actors[cent]["expanded"]:
                    len_ = centroid_actors[cent]["length"]
                    sz_ = centroid_actors[cent]["size"]
                    if len_ >= min_length and sz_ >= min_size:
                        centroid_actors[cent]["actor"].VisibilityOn()
                        cent.VisibilityOff()
                        centroid_actors[cent]["expanded"] = 1
        self.show_m.render()

    # TODO: Move to another class/module
    def __hide(self):
        centroid_actors = self.__clusters_visualizer.centroid_actors
        lengths = self.__clusters_visualizer.lengths
        sizes = self.__clusters_visualizer.sizes
        min_length = np.min(lengths)
        min_size = np.min(sizes)
        for cent in centroid_actors:
            valid_length = centroid_actors[cent]["length"] >= min_length
            valid_size = centroid_actors[cent]["size"] >= min_size
            if self.__hide_centroids:
                if valid_length or valid_size:
                    if centroid_actors[cent]["selected"] == 0:
                        cent.VisibilityOff()
            else:
                if valid_length and valid_size:
                    if centroid_actors[cent]["selected"] == 0:
                        cent.VisibilityOn()
        self.__hide_centroids = not self.__hide_centroids
        self.show_m.render()

    # TODO: Move to another class/module
    def __invert(self):
        centroid_actors = self.__clusters_visualizer.centroid_actors
        cluster_actors = self.__clusters_visualizer.cluster_actors
        lengths = self.__clusters_visualizer.lengths
        sizes = self.__clusters_visualizer.sizes
        min_length = np.min(lengths)
        min_size = np.min(sizes)
        for cent in centroid_actors:
            valid_length = centroid_actors[cent]["length"] >= min_length
            valid_size = centroid_actors[cent]["size"] >= min_size
            if valid_length and valid_size:
                centroid_actors[cent]["selected"] = not centroid_actors[cent][
                    "selected"
                ]
                clus = centroid_actors[cent]["actor"]
                cluster_actors[clus]["selected"] = centroid_actors[cent]["selected"]
        self.show_m.render()

    def __key_press_events(self, obj, event):
        key = obj.GetKeySym()
        if key in ("r", "R") and obj.GetControlKey():
            self._scene.reset_camera_tight()
            self.show_m.render()
        if key in ("o", "O"):
            if self._shortcut_help_visible:
                self._scene.rm(*self.shortcut_panel.actors)
                self._shortcut_help_visible = False
            else:
                self._scene.add(*self.shortcut_panel.actors)
                self._shortcut_help_visible = True
            self.show_m.render()
        # TODO: Move to another class/module
        if self.cluster:
            # retract help panel
            if key in ("o", "O"):
                if self.__help_visible:
                    self._scene.rm(*self.help_panel.actors)
                    self.__help_visible = False
                else:
                    self._scene.add(*self.help_panel.actors)
                    self.__help_visible = True
                self.show_m.render()
            if key in ("a", "A"):
                self.__show_all()
            if key in ("e", "E"):
                self.__expand()
            # hide on/off unselected centroids
            if key in ("h", "H"):
                self.__hide()
            # invert selection
            if key in ("i", "I"):
                self.__invert()
            if key in ("r", "R"):
                self.__reset()
            # save current result
            if key in ("s", "S"):
                self.__save()
            if key in ("y", "Y"):
                self.__new_window()

    # TODO: Move to another class/module
    def __new_window(self):
        cluster_actors = self.__clusters_visualizer.cluster_actors
        tractogram_clusters = self.__clusters_visualizer.tractogram_clusters
        active_streamlines = Streamlines()
        for bundle in cluster_actors.keys():
            if bundle.GetVisibility():
                t = cluster_actors[bundle]["tractogram"]
                c = cluster_actors[bundle]["cluster"]
                indices = tractogram_clusters[t][c]
                active_streamlines.extend(Streamlines(indices))

        # Using the header of the first of the tractograms
        active_sft = StatefulTractogram(
            active_streamlines, self.tractograms[0], Space.RASMM
        )
        hz2 = Horizon(
            tractograms=[active_sft],
            images=self.images,
            cluster=True,
            cluster_thr=self.cluster_thr / 2.0,
            random_colors=self.random_colors,
            length_lt=np.inf,
            length_gt=0,
            clusters_lt=np.inf,
            clusters_gt=0,
            world_coords=True,
            interactive=True,
        )
        ren2 = hz2.build_scene()
        hz2.build_show(ren2)

    # TODO: Move to another class/module
    def __reset(self):
        centroid_actors = self.__clusters_visualizer.centroid_actors
        lengths = self.__clusters_visualizer.lengths
        sizes = self.__clusters_visualizer.sizes
        min_length = np.min(lengths)
        min_size = np.min(sizes)
        for cent in centroid_actors:
            valid_length = centroid_actors[cent]["length"] >= min_length
            valid_size = centroid_actors[cent]["size"] >= min_size
            if valid_length and valid_size:
                centroid_actors[cent]["actor"].VisibilityOff()
                cent.VisibilityOn()
                centroid_actors[cent]["expanded"] = 0
        self.show_m.render()

    # TODO: Move to another class/module
    def __save(self):
        cluster_actors = self.__clusters_visualizer.cluster_actors
        tractogram_clusters = self.__clusters_visualizer.tractogram_clusters
        saving_streamlines = Streamlines()
        for bundle in cluster_actors.keys():
            if bundle.GetVisibility():
                t = cluster_actors[bundle]["tractogram"]
                c = cluster_actors[bundle]["cluster"]
                indices = tractogram_clusters[t][c]
                saving_streamlines.extend(Streamlines(indices))
        logging.info("Saving result in tmp.trk")

        # Using the header of the first of the tractograms
        sft_new = StatefulTractogram(
            saving_streamlines, self.tractograms[0], Space.RASMM
        )
        save_tractogram(sft_new, "tmp.trk", bbox_valid_check=False)
        logging.info("Saved!")

    # TODO: Move to another class/module
    def __show_all(self):
        centroid_actors = self.__clusters_visualizer.centroid_actors
        cluster_actors = self.__clusters_visualizer.cluster_actors
        lengths = self.__clusters_visualizer.lengths
        sizes = self.__clusters_visualizer.sizes
        min_length = np.min(lengths)
        min_size = np.min(sizes)
        if self.__select_all:
            for cent in centroid_actors:
                valid_length = centroid_actors[cent]["length"] >= min_length
                valid_size = centroid_actors[cent]["size"] >= min_size
                if valid_length and valid_size:
                    centroid_actors[cent]["selected"] = 0
                    clus = centroid_actors[cent]["actor"]
                    cluster_actors[clus]["selected"] = centroid_actors[cent]["selected"]
            self.__select_all = False
        else:
            for cent in centroid_actors:
                valid_length = centroid_actors[cent]["length"] >= min_length
                valid_size = centroid_actors[cent]["size"] >= min_size
                if valid_length and valid_size:
                    centroid_actors[cent]["selected"] = 1
                    clus = centroid_actors[cent]["actor"]
                    cluster_actors[clus]["selected"] = centroid_actors[cent]["selected"]
            self.__select_all = True
        self.show_m.render()

    def __win_callback(self, obj, _event):
        if self._win_size != obj.GetSize():
            self._win_size = obj.GetSize()
            if len(self.__tabs) > 0:
                self.__tab_mgr.reposition(self._win_size)
            panel_size = self.shortcut_panel._get_size()
            self.shortcut_panel._set_position(
                (5, self._win_size[1] - panel_size[1] - 5)
            )
            if self.cluster:
                panel_size = self.help_panel._get_size()
                new_pos = np.array(self._win_size) - panel_size - 5
                self.help_panel._set_position(new_pos)

    def build_scene(self):
        self.mem = GlobalHorizon()
        scene = window.Scene()
        scene.background(self.bg_color)
        return scene

    def _show_force_render(self, _element):
        """
        Callback function for lower level elements to force render.
        """
        self.show_m.render()

    def _update_actors(self, actors):
        """Update actors in the scene. It essentially brings them forward in
        the stack.

        Parameters
        ----------
        actors : list
            list of FURY actors.
        """
        self._scene.rm(*actors)
        self._scene.add(*actors)

    def build_show(self, scene):
        self._scene = scene
        self._win_size = self._scene.GetSize()

        title = f"Horizon {horizon_version}"
        self.show_m = window.ShowManager(
            scene=scene,
            title=title,
            size=(1920, 1080),
            reset_camera=False,
            order_transparent=self.order_transparent,
        )

        if len(self.tractograms) > 0:
            if self.cluster:
                self.__clusters_visualizer = ClustersVisualizer(
                    self.show_m, scene, self.tractograms
                )

            color_ind = 0

            for t, sft in enumerate(self.tractograms):
                streamlines = sft.streamlines

                if "tracts" in self.random_colors:
                    colors = next(self.color_gen)
                else:
                    colors = None

                if not self.world_coords:
                    # TODO: Get affine from a StatefullTractogram
                    raise ValueError(
                        "Currently native coordinates are not supported for "
                        "streamlines."
                    )

                if self.cluster:
                    self.__clusters_visualizer.add_cluster_actors(
                        t, streamlines, self.cluster_thr, colors
                    )
                else:
                    if self.buan:
                        colors = self.buan_colors[color_ind]

                    streamline_actor = actor.line(streamlines, colors=colors)
                    streamline_actor.GetProperty().SetEdgeVisibility(1)
                    streamline_actor.GetProperty().SetRenderLinesAsTubes(1)
                    streamline_actor.GetProperty().SetLineWidth(6)
                    streamline_actor.GetProperty().SetOpacity(1)
                    scene.add(streamline_actor)

                color_ind += 1

            if self.cluster:
                # Information panel
                # It will be changed once all the elements wrapped in horizon
                # elements.
                text_block = build_label(HELP_MESSAGE, font_size=18)

                self.help_panel = ui.Panel2D(
                    size=(300, 200),
                    position=(1615, 875),
                    color=(0.8, 0.8, 1.0),
                    opacity=0.2,
                    align="left",
                )

                self.help_panel.add_element(text_block.obj, coords=(0.02, 0.01))
                scene.add(self.help_panel)
                self.__tabs.append(
                    ClustersTab(self.__clusters_visualizer, self.cluster_thr)
                )

        text_block = build_label(SHORTCUT_MESSAGE, font_size=18)

        self.shortcut_panel = ui.Panel2D(
            size=(260, 115),
            position=(5, 960),
            color=(0.8, 0.8, 1.0),
            opacity=0.2,
            align="left",
        )
        self.shortcut_panel.add_element(text_block.obj, coords=(0.02, 0.01))
        scene.add(self.shortcut_panel)
        sync_slices = sync_vol = False
        self.images = check_img_dtype(self.images)
        if len(self.images) > 0:
            if self.__roi_images:
                roi_color = self.__roi_colors
            roi_actors = []
            img_count = 0
            sync_slices, sync_vol = check_img_shapes(self.images)
            for img in self.images:
                title = f"Image {img_count + 1}"
                data, affine, fname = unpack_image(img)
                self.vox2ras = affine
                if self.__roi_images:
                    if "rois" in self.random_colors:
                        roi_color = next(self.color_gen)
                    roi_actor = actor.contour_from_roi(
                        data, affine=affine, color=roi_color
                    )
                    scene.add(roi_actor)
                    roi_actors.append(roi_actor)
                else:
                    slices_viz = SlicesVisualizer(
                        self.show_m.iren,
                        scene,
                        data,
                        affine=affine,
                        world_coords=self.world_coords,
                        rgb=self.rgb,
                        fname=fname,
                    )
                    self.__tabs.append(
                        SlicesTab(
                            slices_viz,
                            title,
                            fname,
                            force_render=self._show_force_render,
                        )
                    )
                    img_count += 1

            if len(roi_actors) > 0:
                self.__tabs.append(ROIsTab(roi_actors))

        sync_peaks = False
        if len(self.pams) > 0:
            if self.images:
                sync_peaks = check_peak_size(
                    self.pams,
                    ref_img_shape=self.images[0][0].shape[:3],
                    sync_imgs=sync_slices,
                )
            else:
                sync_peaks = check_peak_size(self.pams)
            for pam in self.pams:
                peak_viz = PeaksVisualizer(
                    (pam.peak_dirs, pam.affine), self.world_coords
                )
                scene.add(peak_viz.actors[0])
                self.__tabs.append(PeaksTab(peak_viz.actors[0]))

        if len(self._surfaces) > 0:
            for idx, surface in enumerate(self._surfaces):
                try:
                    vertices, faces, fname = unpack_surface(surface)
                except ValueError as e:
                    warn(str(e), stacklevel=2)
                    continue
                color = next(self.color_gen)
                title = f"Surface {idx + 1}"
                surf_viz = SurfaceVisualizer((vertices, faces), scene, color)
                surf_tab = SurfaceTab(surf_viz, title, fname)
                self.__tabs.append(surf_tab)

        if len(self.__tabs) > 0:
            self.__tab_mgr = TabManager(
                tabs=self.__tabs,
                win_size=scene.GetSize(),
                on_tab_changed=self._update_actors,
                add_to_scene=self._scene.add,
                remove_from_scene=self._scene.rm,
                sync_slices=sync_slices,
                sync_volumes=sync_vol,
                sync_peaks=sync_peaks,
            )

            scene.add(self.__tab_mgr.tab_ui)
            self.__tab_mgr.handle_text_overflows()
            self.__tabs[-1].on_tab_selected()

        self.show_m.initialize()

        options = [
            r"un\hide centroids",
            "invert selection",
            r"un\select all",
            "expand clusters",
            "collapse clusters",
            "save streamlines",
            "recluster",
        ]
        listbox = ui.ListBox2D(
            values=options,
            position=(10, 300),
            size=(200, 270),
            multiselection=False,
            font_size=18,
        )

        def display_element():
            action = listbox.selected[0]
            if action == r"un\hide centroids":
                self.__hide()
            if action == "invert selection":
                self.__invert()
            if action == r"un\select all":
                self.__show_all()
            if action == "expand clusters":
                self.__expand()
            if action == "collapse clusters":
                self.__reset()
            if action == "save streamlines":
                self.__save()
            if action == "recluster":
                self.__new_window()

        listbox.on_change = display_element
        listbox.panel.opacity = 0.2
        listbox.set_visibility(0)

        self.show_m.scene.add(listbox)

        def left_click_centroid_callback(obj, event):
            self.cea[obj]["selected"] = not self.cea[obj]["selected"]
            self.cla[self.cea[obj]["cluster_actor"]]["selected"] = self.cea[obj][
                "selected"
            ]
            self.show_m.render()

        def right_click_centroid_callback(obj, event):
            for lactor in listbox._get_actors():
                lactor.SetVisibility(not lactor.GetVisibility())

            listbox.scroll_bar.set_visibility(False)
            self.show_m.render()

        def left_click_cluster_callback(obj, event):
            if self.cla[obj]["selected"]:
                self.cla[obj]["centroid_actor"].VisibilityOn()
                ca = self.cla[obj]["centroid_actor"]
                self.cea[ca]["selected"] = 0
                obj.VisibilityOff()
                self.cea[ca]["expanded"] = 0

            self.show_m.render()

        def right_click_cluster_callback(obj, event):
            logging.info("Cluster Area Selected")
            self.show_m.render()

        for cl in self.cla:
            cl.AddObserver("LeftButtonPressEvent", left_click_cluster_callback, 1.0)
            cl.AddObserver("RightButtonPressEvent", right_click_cluster_callback, 1.0)
            self.cla[cl]["centroid_actor"].AddObserver(
                "LeftButtonPressEvent", left_click_centroid_callback, 1.0
            )
            self.cla[cl]["centroid_actor"].AddObserver(
                "RightButtonPressEvent", right_click_centroid_callback, 1.0
            )

        self.mem.window_timer_cnt = 0

        def timer_callback(obj, event):
            self.mem.window_timer_cnt += 1
            # TODO possibly add automatic rotation option
            # self.show_m.scene.azimuth(0.01 * self.mem.window_timer_cnt)
            # self.show_m.render()

        scene.reset_camera()
        scene.zoom(1.5)
        scene.reset_clipping_range()

        if self.interactive:
            self.show_m.add_window_callback(self.__win_callback)
            self.show_m.add_timer_callback(True, 200, timer_callback)
            self.show_m.iren.AddObserver("KeyPressEvent", self.__key_press_events)

            if self.return_showm:
                return self.show_m

            if self.recorded_events is None:
                self.show_m.render()
                self.show_m.start()

            else:
                # set to True if event recorded file needs updating
                recording = False
                recording_filename = self.recorded_events

                if recording:
                    self.show_m.record_events_to_file(recording_filename)
                else:
                    self.show_m.play_events_from_file(recording_filename)

        else:
            window.record(
                scene=scene, out_path=self.out_png, size=(1200, 900), reset_camera=False
            )


@warning_for_keywords()
def horizon(
    *,
    tractograms=None,
    images=None,
    pams=None,
    surfaces=None,
    cluster=False,
    rgb=False,
    cluster_thr=15.0,
    random_colors=None,
    bg_color=(0, 0, 0),
    order_transparent=True,
    length_gt=0,
    length_lt=1000,
    clusters_gt=0,
    clusters_lt=10000,
    world_coords=True,
    interactive=True,
    buan=False,
    buan_colors=None,
    roi_images=False,
    roi_colors=(1, 0, 0),
    out_png="tmp.png",
    recorded_events=None,
    return_showm=False,
):
    """Interactive medical visualization - Invert the Horizon!

    See :footcite:p:`Garyfallidis2019` for further details about Horizon.

    Parameters
    ----------
    tractograms : sequence of StatefulTractograms
        StatefulTractograms are used for making sure that the coordinate
        systems are correct
    images : sequence of tuples
        Each tuple contains data and affine
    pams : sequence of PeakAndMetrics
        Contains peak directions and spherical harmonic coefficients
    surfaces : sequence of tuples
        Each tuple contains vertices and faces
    cluster : bool
        Enable QuickBundlesX clustering
    rgb: bool, optional
        Enable the color image.
    cluster_thr : float
        Distance threshold used for clustering. Default value 15.0 for
        small animal data you may need to use something smaller such
        as 2.0. The threshold is in mm. For this parameter to be active
        ``cluster`` should be enabled.
    random_colors : string
        Given multiple tractograms and/or ROIs then each tractogram and/or
        ROI will be shown with different color. If no value is provided both
        the tractograms and the ROIs will have a different random color
        generated from a distinguishable colormap. If the effect should only be
        applied to one of the 2 objects, then use the options 'tracts' and
        'rois' for the tractograms and the ROIs respectively.
    bg_color : ndarray or list or tuple
        Define the background color of the scene. Default is black (0, 0, 0)
    order_transparent : bool
        Default True. Use depth peeling to sort transparent objects.
        If True also enables anti-aliasing.
    length_gt : float
        Clusters with average length greater than ``length_gt`` amount
        in mm will be shown.
    length_lt : float
        Clusters with average length less than ``length_lt`` amount in mm
        will be shown.
    clusters_gt : int
        Clusters with size greater than ``clusters_gt`` will be shown.
    clusters_lt : int
        Clusters with size less than ``clusters_lt`` will be shown.
    world_coords : bool
        Show data in their world coordinates (not native voxel coordinates)
        Default True.
    interactive : bool
        Allow user interaction. If False then Horizon goes on stealth mode
        and just saves pictures.
    buan : bool, optional
        Enables BUAN framework visualization. Default is False.
    buan_colors : list, optional
        List of colors for bundles.
    roi_images : bool, optional
        Displays binary images as contours. Default is False.
    roi_colors : ndarray or list or tuple, optional
        Define the color of the roi images. Default is red (1, 0, 0)
    out_png : string
        Filename of saved picture.
    recorded_events : string
        File path to replay recorded events
    return_showm : bool
        Return ShowManager object. Used only at Python level. Can be used
        for extending Horizon's capabilities externally and for testing
        purposes.

    References
    ----------
    .. footbibliography::
    """

    hz = Horizon(
        tractograms=tractograms,
        images=images,
        pams=pams,
        surfaces=surfaces,
        cluster=cluster,
        rgb=rgb,
        cluster_thr=cluster_thr,
        random_colors=random_colors,
        length_gt=length_gt,
        length_lt=length_lt,
        clusters_gt=clusters_gt,
        clusters_lt=clusters_lt,
        world_coords=world_coords,
        interactive=interactive,
        out_png=out_png,
        recorded_events=recorded_events,
        return_showm=return_showm,
        bg_color=bg_color,
        order_transparent=order_transparent,
        buan=buan,
        buan_colors=buan_colors,
        roi_images=roi_images,
        roi_colors=roi_colors,
    )

    scene = hz.build_scene()

    if return_showm:
        return hz.build_show(scene)
    hz.build_show(scene)



from os.path import join as pjoin
import sys
from tempfile import TemporaryDirectory

import numpy.testing as npt

from dipy.workflows.base import (
    IntrospectiveArgumentParser,
    add_default_args_to_docstring,
    none_or_dtype,
)
from dipy.workflows.flow_runner import run_flow
from dipy.workflows.tests.workflow_tests_utils import (
    DummyCombinedWorkflow,
    DummyFlow,
    DummyVariableTypeErrorWorkflow,
    DummyVariableTypeWorkflow,
    DummyWorkflow1,
    DummyWorkflowOptionalStr,
)


def test_none_or_dtype():
    # test None
    for typ in [int, float, str, tuple, list]:
        dec = none_or_dtype(typ)
        npt.assert_equal(dec("None"), "None")
        npt.assert_equal(dec("none"), "None")

    dec = none_or_dtype(int)
    npt.assert_raises(ValueError, dec, "my value")
    npt.assert_equal(dec(4), 4)

    dec = none_or_dtype(str)
    npt.assert_equal(dec(4), "4")
    npt.assert_equal(dec("my value"), "my value")
    npt.assert_equal(dec([4]), "[4]")

    dec = none_or_dtype(float)
    npt.assert_raises(ValueError, dec, "my value")
    for val in [
        (4,),
        [
            4,
        ],
    ]:
        npt.assert_raises(TypeError, dec, val)
    npt.assert_equal(dec(True), 1.0)
    npt.assert_equal(dec(4), 4.0)
    npt.assert_equal(dec(4.0), 4.0)

    dec = none_or_dtype(bool)
    dec = none_or_dtype(tuple)


def test_variable_type():
    with TemporaryDirectory() as out_dir:
        open(pjoin(out_dir, "test"), "w").close()
        open(pjoin(out_dir, "test1"), "w").close()
        open(pjoin(out_dir, "test2"), "w").close()

        sys.argv = [sys.argv[0]]
        pos_results = [
            pjoin(out_dir, "test"),
            pjoin(out_dir, "test1"),
            pjoin(out_dir, "test2"),
            12,
        ]
        inputs = inputs_from_results(pos_results)
        sys.argv.extend(inputs)
        dcwf = DummyVariableTypeWorkflow()
        _, positional_res, positional_res2 = run_flow(dcwf)
        npt.assert_equal(positional_res2, 12)

        for k, v in zip(positional_res, pos_results[:-1]):
            npt.assert_equal(k, v)

        dcwf = DummyVariableTypeErrorWorkflow()
        npt.assert_raises(ValueError, run_flow, dcwf)


def test_iap():
    sys.argv = [sys.argv[0]]
    pos_keys = [
        "positional_str",
        "positional_bool",
        "positional_int",
        "positional_float",
    ]

    opt_keys = [
        "optional_str",
        "optional_bool",
        "optional_int",
        "optional_float",
        "optional_int_2",
        "optional_float_2",
    ]

    pos_results = ["test", 0, 10, 10.2]
    opt_results = ["opt_test", True, 20, 20.2, None, None]

    inputs = inputs_from_results(opt_results, opt_keys, optional=True)
    inputs.extend(inputs_from_results(pos_results))

    sys.argv.extend(inputs)
    parser = IntrospectiveArgumentParser()
    dummy_flow = DummyFlow()
    parser.add_workflow(dummy_flow)
    args = parser.get_flow_args()
    all_keys = pos_keys + opt_keys
    all_results = pos_results + opt_results

    # Test if types and order are respected
    for k, v in zip(all_keys, all_results):
        npt.assert_equal(args[k], v)

    # Test if **args really fits dummy_flow's arguments
    return_values = dummy_flow.run(**args)
    npt.assert_array_equal(return_values, all_results + [2.0])


def test_optional_str():
    # Test optional and variable str argument exists but does not have a value
    sys.argv = [sys.argv[0]]
    inputs = ["--optional_str_1"]
    sys.argv.extend(inputs)
    parser = IntrospectiveArgumentParser()
    dummy_flow = DummyWorkflowOptionalStr()
    parser.add_workflow(dummy_flow)
    args = parser.get_flow_args()
    all_keys = ["optional_str_1"]
    all_results = [[]]
    # Test if types and order are respected
    for k, v in zip(all_keys, all_results):
        npt.assert_equal(args[k], v)
    # Test if **args really fits dummy_flow's arguments
    return_values = dummy_flow.run(**args)
    npt.assert_equal(return_values, all_results + ["default"])

    # Test optional and variable str argument exists and has a value
    sys.argv = [sys.argv[0]]
    inputs = ["--optional_str_1", "test"]
    sys.argv.extend(inputs)
    parser = IntrospectiveArgumentParser()
    dummy_flow = DummyWorkflowOptionalStr()
    parser.add_workflow(dummy_flow)
    args = parser.get_flow_args()
    all_keys = ["optional_str_1"]
    all_results = [["test"]]
    # Test if types and order are respected
    for k, v in zip(all_keys, all_results):
        npt.assert_equal(args[k], v)
    # Test if **args really fits dummy_flow's arguments
    return_values = dummy_flow.run(**args)
    npt.assert_equal(return_values, all_results + ["default"])

    # Test optional str empty arguments
    sys.argv = [sys.argv[0]]
    inputs = ["--optional_str_2"]
    sys.argv.extend(inputs)
    parser = IntrospectiveArgumentParser()
    dummy_flow = DummyWorkflowOptionalStr()
    parser.add_workflow(dummy_flow)
    with npt.assert_raises(SystemExit) as cm:
        parser.get_flow_args()
    npt.assert_equal(cm.exception.code, 2)


def test_iap_epilog_and_description():
    parser = IntrospectiveArgumentParser()
    dummy_flow = DummyWorkflow1()
    parser.add_workflow(dummy_flow)
    assert "dummy references" in parser.epilog
    assert "Workflow used to test combined" in parser.description


def test_flow_runner():
    old_argv = sys.argv
    sys.argv = [sys.argv[0]]

    opt_keys = [
        "param_combined",
        "dwf1.param1",
        "dwf2.param2",
        "force",
        "out_strat",
        "mix_names",
    ]

    pos_results = ["dipy.txt"]
    opt_results = [30, 10, 20, True, "absolute", True]

    inputs = inputs_from_results(opt_results, opt_keys, optional=True)
    inputs.extend(inputs_from_results(pos_results))

    sys.argv.extend(inputs)

    dcwf = DummyCombinedWorkflow()
    param1, param2, combined = run_flow(dcwf)

    # generic flow params
    assert dcwf._force_overwrite
    assert dcwf._output_strategy == "absolute"
    assert dcwf._mix_names

    # sub flow params
    assert param1 == 10
    assert param2 == 20

    # parent flow param
    assert combined == 30

    sys.argv = old_argv


def inputs_from_results(results, keys=None, optional=False):
    prefix = "--"
    inputs = []
    for idx, result in enumerate(results):
        if keys is not None:
            inputs.append(prefix + keys[idx])
        if optional and str(result) in ["True", "False"]:
            continue
        inputs.append(str(result))

    return inputs


def test_add_default_args_to_docstring():
    # --- Test 1: Adds default values correctly ---
    def func1(a, b=10, c="hello"):
        """Sample function."""
        pass

    npds1 = {"Parameters": [("a", "int", []), ("b", "int", []), ("c", "str", [])]}
    add_default_args_to_docstring(npds1, func1)
    assert npds1["Parameters"][1][2] == ["(default: 10)"]
    assert npds1["Parameters"][2][2] == ["(default: hello)"]

    # --- Test 2: Ignores non-default parameters ---
    def func2(x, y):
        """Function without defaults."""
        pass

    npds2 = {"Parameters": [("x", "int", []), ("y", "int", [])]}

    add_default_args_to_docstring(npds2, func2)

    assert npds2["Parameters"][0][2] == []
    assert npds2["Parameters"][1][2] == []

    # --- Test 3: Handles 'out_dir' special case ---
    def func3(out_dir="."):
        """Handles out_dir default properly."""
        pass

    npds3 = {"Parameters": [("out_dir", "str", [])]}

    add_default_args_to_docstring(npds3, func3)

    assert npds3["Parameters"][0][2] == ["(default:  current directory)"]

    # --- Test 4: Handles empty docstring parameters ---
    def func4(a=42):
        """Function with an empty parameter list in docstring."""
        pass

    npds4 = {"Parameters": []}  # No parameters in docstring

    add_default_args_to_docstring(npds4, func4)

    assert npds4["Parameters"] == []  # Should remain unchanged

    # --- Test 5: Missing parameter in docstring ---
    def func5(a=42, b="test"):
        """Function with some missing parameters in docstring."""
        pass

    npds5 = {
        "Parameters": [
            ("a", "int", [])  # 'b' is missing
        ]
    }

    add_default_args_to_docstring(npds5, func5)

    assert npds5["Parameters"][0][2] == ["(default: 42)"]  # Only 'a' updated
    assert len(npds5["Parameters"]) == 1  # 'b' should not be added automatically

    # --- Test 6: Handles multi-line parameter descriptions ---
    def func6(alpha=0.5, beta=0.3):
        """Function with multi-line parameter descriptions.

        Parameters
        ----------
        alpha : float
            Learning rate parameter.
            This value controls step size.
        beta : float
            Another parameter with
            multiple lines of description.
        """
        pass

    npds6 = {
        "Parameters": [
            (
                "alpha",
                "float",
                ["Learning rate parameter.", "This value controls step size."],
            ),
            (
                "beta",
                "float",
                ["Another parameter with", "multiple lines of description."],
            ),
        ]
    }

    add_default_args_to_docstring(npds6, func6)

    assert npds6["Parameters"][0][2] == [
        "Learning rate parameter.",
        "This value controls step size.",
        "(default: 0.5)",
    ]
    assert npds6["Parameters"][1][2] == [
        "Another parameter with",
        "multiple lines of description.",
        "(default: 0.3)",
    ]



"""
===================================================
Reconstruction with the Sparse Fascicle Model (SFM)
===================================================

In this example, we will use the Sparse Fascicle Model (SFM)
:footcite:p:`Rokem2015`, to reconstruct the fiber Orientation Distribution
Function (fODF) in every voxel.

First, we import the modules we will use in this example:
"""

from dipy.core.gradients import gradient_table
import dipy.data as dpd
import dipy.direction.peaks as dpp
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
from dipy.reconst.csdeconv import auto_response_ssst
import dipy.reconst.sfm as sfm
from dipy.viz import actor, window

###############################################################################
# For the purpose of this example, we will use the Stanford HARDI dataset (150
# directions, single b-value of 2000 $s/mm^2$) that can be automatically
# downloaded. If you have not yet downloaded this data-set in one of the other
# examples, you will need to be connected to the internet the first time you
# run this example. The data will be stored for subsequent runs, and for use
# with other examples.

hardi_fname, hardi_bval_fname, hardi_bvec_fname = dpd.get_fnames(name="stanford_hardi")
data, affine = load_nifti(hardi_fname)

bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

# Enables/disables interactive visualization
interactive = False

###############################################################################
# Reconstruction of the fiber ODF in each voxel guides subsequent tracking
# steps. Here, the model is the Sparse Fascicle Model, described in
# :footcite:p:`Rokem2015`. This model reconstructs the diffusion signal as a
# combination of the signals from different fascicles. This model can be written
# as:
#
# .. math::
#
#     y = X\beta
#
# Where $y$ is the signal and $\beta$ are weights on different points in the
# sphere. The columns of the design matrix, $X$ are the signals in each point
# in the measurement that would be predicted if there was a fascicle oriented
# in the direction represented by that column. Typically, the signal used for
# this kernel will be a prolate tensor with axial diffusivity 3-5 times higher
# than its radial diffusivity. The exact numbers can also be estimated from
# examining parts of the brain in which there is known to be only one fascicle
# (e.g. in corpus callosum).
#
# Sparsity constraints on the fiber ODF ($\beta$) are set through the Elastic
# Net algorithm :footcite:p:`Zou2005`.
#
# Elastic Net optimizes the following cost function:
#
# .. math::
#
#     \sum_{i=1}^{n}{(y_i - \hat{y}_i)^2} + \alpha (\lambda \sum_{j=1}^{m}{w_j}+(1-\lambda) \sum_{j=1}^{m}{w^2_j}  # noqa: E501
#
# where $\hat{y}$ is the signal predicted for a particular setting of $\beta$,
# such that the left part of this expression is the squared loss function;
# $\alpha$ is a parameter that sets the balance between the squared loss on
# the data, and the regularization constraints. The regularization parameter
# $\lambda$ sets the `l1_ratio`, which controls the balance between L1-sparsity
# (low sum of weights), and low L2-sparsity (low sum-of-squares of the
# weights).
#
# Just like Constrained Spherical Deconvolution (see :ref:`reconst-csd`), the
# SFM requires the definition of a response function. We'll take advantage of
# the automated algorithm in the :mod:`csdeconv` module to find this response
# function:

response, ratio = auto_response_ssst(gtab, data, roi_radii=10, fa_thr=0.7)

###############################################################################
# The ``response`` return value contains two entries. The first is an array
# with the eigenvalues of the response function and the second is the average
# S0 for this response.
#
# It is a very good practice to always validate the result of
# ``auto_response_ssst``. For, this purpose we can print it and have a look
# at its values.

print(response)

###############################################################################
# We initialize an SFM model object, using these values. We will use the
# default sphere (362  vertices, symmetrically distributed on the surface of
# the sphere), as a set of putative fascicle directions that are considered
# in the model

sphere = dpd.get_sphere()
sf_model = sfm.SparseFascicleModel(
    gtab, sphere=sphere, l1_ratio=0.5, alpha=0.001, response=response[0]
)

###############################################################################
# For the purpose of the example, we will consider a small volume of data
# containing parts of the corpus callosum and of the centrum semiovale

data_small = data[20:50, 55:85, 38:39]

###############################################################################
# Fitting the model to this small volume of data, we calculate the ODF of this
# model on the sphere, and plot it.

sf_fit = sf_model.fit(data_small)
sf_odf = sf_fit.odf(sphere)

fodf_spheres = actor.odf_slicer(sf_odf, sphere=sphere, scale=0.8, colormap="plasma")

scene = window.Scene()
scene.add(fodf_spheres)

window.record(scene=scene, out_path="sf_odfs.png", size=(1000, 1000))
if interactive:
    window.show(scene)

###############################################################################
# We can extract the peaks from the ODF, and plot these as well

sf_peaks = dpp.peaks_from_model(
    sf_model,
    data_small,
    sphere,
    relative_peak_threshold=0.5,
    min_separation_angle=25,
    return_sh=False,
)


scene.clear()
fodf_peaks = actor.peak_slicer(sf_peaks.peak_dirs, peaks_values=sf_peaks.peak_values)
scene.add(fodf_peaks)

window.record(scene=scene, out_path="sf_peaks.png", size=(1000, 1000))
if interactive:
    window.show(scene)

###############################################################################
# Finally, we plot both the peaks and the ODFs, overlaid:

fodf_spheres.GetProperty().SetOpacity(0.4)
scene.add(fodf_spheres)

window.record(scene=scene, out_path="sf_both.png", size=(1000, 1000))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# SFM Peaks and ODFs.
#
#
# To see how to use this information in tracking, proceed to
# :ref:`sphx_glr_examples_built_fiber_tracking_tracking_sfm.py`.
#
# References
# ----------
#
# .. footbibliography::
#



#!/usr/bin/python
"""Classes and functions for fitting the correlation tensor model"""

import numpy as np

from dipy.core.onetime import auto_attr
from dipy.reconst.base import ReconstModel
from dipy.reconst.dki import (
    DiffusionKurtosisFit,
)
from dipy.reconst.dti import (
    MIN_POSITIVE_SIGNAL,
    decompose_tensor,
    from_lower_triangular,
    lower_triangular,
)
from dipy.reconst.multi_voxel import multi_voxel_fit
from dipy.reconst.utils import cti_design_matrix as design_matrix
from dipy.testing.decorators import warning_for_keywords


def from_qte_to_cti(C):
    """
    Rescales the qte C elements to the C elements used in CTI.

    Parameters
    ----------
    C: array(..., 21)
        Twenty-one elements of the covariance tensor in voigt notation plus
        some extra scaling factors.

    Returns
    -------
    ccti: array(..., 21)
        Covariance Tensor Elements with no hidden factors.
    """
    const = np.sqrt(2)
    ccti = np.zeros((21, 1))
    ccti[0] = C[0]
    ccti[1] = C[1]
    ccti[2] = C[2]
    ccti[3] = C[3] / const
    ccti[4] = C[4] / const
    ccti[5] = C[5] / const
    ccti[6] = C[6] / 2
    ccti[7] = C[7] / 2
    ccti[8] = C[8] / 2
    ccti[9] = C[9] / 2
    ccti[10] = C[10] / 2
    ccti[11] = C[11] / 2
    ccti[12] = C[12] / 2
    ccti[13] = C[13] / 2
    ccti[14] = C[14] / 2
    ccti[15] = C[15] / 2
    ccti[16] = C[16] / 2
    ccti[17] = C[17] / 2
    ccti[18] = C[18] / (2 * const)
    ccti[19] = C[19] / (2 * const)
    ccti[20] = C[20] / (2 * const)
    return ccti


def multi_gaussian_k_from_c(ccti, MD):
    """
    Computes the multiple Gaussian diffusion kurtosis tensor from the
    covariance tensor.

    Parameters
    ----------
    ccti: array(..., 21)
        Covariance Tensor Elements with no hidden factors.
    MD : ndarray
        Mean Diffusivity (MD) of a diffusion tensor.

    Returns
    -------
    K: array (..., 15)
        Fifteen elements of the kurtosis tensor
    """
    K = np.zeros((15, 1))
    K[0] = 3 * ccti[0] / (MD**2)
    K[1] = 3 * ccti[1] / (MD**2)
    K[2] = 3 * ccti[2] / (MD**2)
    K[3] = 3 * ccti[8] / (MD**2)
    K[4] = 3 * ccti[7] / (MD**2)
    K[5] = 3 * ccti[11] / (MD**2)
    K[6] = 3 * ccti[9] / (MD**2)
    K[7] = 3 * ccti[13] / (MD**2)
    K[8] = 3 * ccti[12] / (MD**2)
    K[9] = (ccti[5] + 2 * ccti[17]) / (MD**2)
    K[10] = (ccti[4] + 2 * ccti[16]) / (MD**2)
    K[11] = (ccti[3] + 2 * ccti[15]) / (MD**2)
    K[12] = (ccti[6] + 2 * ccti[19]) / (MD**2)
    K[13] = (ccti[10] + 2 * ccti[20]) / (MD**2)
    K[14] = (ccti[14] + 2 * ccti[18]) / (MD**2)
    return K


def split_cti_params(cti_params):
    r"""Splits CTI params into DTI, DKI, CTI portions.

    Extract the diffusion tensor eigenvalues, the diffusion tensor
    eigenvector matrix, and the 21 independent elements of the covariance
    tensor, and the 15 independent elements of the kurtosis tensor from the
    model parameters estimated from the CTI model

    Parameters
    ----------
    cti_params: numpy.ndarray (..., 48)
        All parameters estimated from the correlation tensor model.
        Parameters are ordered as follows:

            1. Three diffusion tensor's eigenvalues
            2. Three lines of the eigenvector matrix each containing the
               first, second and third coordinates of the eigenvector
            3. Fifteen elements of the kurtosis tensor
            4. Twenty-One elements of the covariance tensor

    Returns
    -------
    evals : array (..., 3)
        Eigenvalues from eigen decomposition of the tensor.
    evecs : array (..., 3)
        Associated eigenvectors from eigen decomposition of the tensor.
        Eigenvectors are columnar (e.g. evecs[:,j] is associated with
        evals[j])
    kt : array (..., 15)
        Fifteen elements of the kurtosis tensor
    ct: array(..., 21)
        Twenty-one elements of the covariance tensor
    """
    evals = cti_params[..., :3]
    evecs = cti_params[..., 3:12].reshape(cti_params.shape[:-1] + (3, 3))
    kt = cti_params[..., 12:27]
    ct = cti_params[..., 27:48]
    return evals, evecs, kt, ct


@warning_for_keywords()
def cti_prediction(cti_params, gtab1, gtab2, *, S0=1):
    """Predict a signal given correlation tensor imaging parameters.

    Parameters
    ----------
    cti_params: numpy.ndarray (..., 48)
        All parameters estimated from the correlation tensor model.
        Parameters are ordered as follows:

            1. Three diffusion tensor's eigenvalues
            2. Three lines of the eigenvector matrix each containing the
               first, second and third coordinates of the eigenvector
            3. Fifteen elements of the kurtosis tensor
            4. Twenty-One elements of the covariance tensor

    gtab1: dipy.core.gradients.GradientTable
        A GradientTable class instance for first DDE diffusion epoch

    gtab2: dipy.core.gradients.GradientTable
        A GradientTable class instance for second DDE diffusion epoch

    S0 : float or ndarray, optional
        The non diffusion-weighted signal in every voxel, or across all
        voxels. Default: 1

    Returns
    -------
    S : ndarray
        Simulated signal based on the CTI model

    """
    evals, evecs, kt, ct = split_cti_params(cti_params)
    A = design_matrix(gtab1, gtab2)
    fevals = evals.reshape((-1, evals.shape[-1]))
    fevecs = evecs.reshape((-1,) + evecs.shape[-2:])
    fct = ct.reshape((-1, ct.shape[-1]))
    fkt = kt.reshape((-1, kt.shape[-1]))
    pred_sig = np.zeros((len(fevals), len(gtab1.bvals)))

    if isinstance(S0, np.ndarray):
        S0_vol = np.reshape(S0, (len(fevals)))
    else:
        S0_vol = S0
    for v in range(len(pred_sig)):
        DT = np.dot(np.dot(fevecs[v], np.diag(fevals[v])), fevecs[v].T)
        dt = lower_triangular(DT)
        MD = (dt[0] + dt[2] + dt[5]) / 3
        if isinstance(S0_vol, np.ndarray):
            this_S0 = S0_vol[v]
        else:
            this_S0 = S0_vol
        X = np.concatenate(
            (dt, fkt[v] * MD * MD, fct[v], np.array([-np.log(this_S0)])), axis=0
        )
        pred_sig[v] = np.exp(np.dot(A, X))
    pred_sig = pred_sig.reshape(cti_params.shape[:-1] + (pred_sig.shape[-1],))

    return pred_sig


class CorrelationTensorModel(ReconstModel):
    """Class for the Correlation Tensor Model"""

    def __init__(self, gtab1, gtab2, *args, fit_method="WLS", **kwargs):
        """Correlation Tensor Imaging Model.

        See :footcite:p:`NetoHenriques2020` for further details about the model.

        Parameters
        ----------
        gtab1: dipy.core.gradients.GradientTable
            A GradientTable class instance for first DDE diffusion epoch
        gtab2: dipy.core.gradients.GradientTable
            A GradientTable class instance for second DDE diffusion epoch
        fit_method : str or callable, optional
            Fitting method.
        *args
            Variable length argument list passed to the :func:`fit` method.
        **kwargs
            Arbitrary keyword arguments passed to the :func:`fit` method.

        References
        ----------
        .. footbibliography::
        """
        self.gtab1 = gtab1
        self.gtab2 = gtab2
        self.args = args
        self.kwargs = kwargs
        self.common_fit_method = not callable(fit_method)

        if self.common_fit_method:
            try:
                self.fit_method = common_fit_methods[fit_method]
            except KeyError as e:
                msg = '"' + str(fit_method) + '" is not a known fit method. The'
                msg += " fit method should either be a function or one of the"
                msg += " common fit methods."
                raise ValueError(msg) from e

        self.args = args
        self.kwargs = kwargs

        self.min_signal = self.kwargs.pop("min_signal", None)
        if self.min_signal is None:
            self.min_signal = MIN_POSITIVE_SIGNAL
        elif self.min_signal <= 0:
            msg = "The `min_signal` key-word argument needs to be strictly"
            msg += " positive."
            raise ValueError(msg)

        self.design_matrix = design_matrix(self.gtab1, self.gtab2)
        self.inverse_design_matrix = np.linalg.pinv(self.design_matrix)

        tol = 1e-6
        self.min_diffusivity = tol / -self.design_matrix.min()
        self.weights = fit_method in {"WLS", "WLLS", "UWLLS"}

    @multi_voxel_fit
    @warning_for_keywords()
    def fit(self, data, *, mask=None):
        """Fit method of the CTI model class.

        Parameters
        ----------
        data : array
            4D array of dMRI data.

        mask : array, optional
            A boolean array of the same shape as data.shape[-1]. It
            designates which coordinates in the data should be analyzed.
        """
        data_thres = np.maximum(data, self.min_signal)
        params = self.fit_method(
            self.design_matrix,
            data_thres,
            self.inverse_design_matrix,
            weights=self.weights,
            *self.args,
            **self.kwargs,
        )

        return CorrelationTensorFit(self, params)

    @warning_for_keywords()
    def predict(self, cti_params, *, S0=1):
        """Predict a signal for the CTI model class instance given parameters

        Parameters
        ----------
        cti_params: numpy.ndarray (..., 48)
            All parameters estimated from the correlation tensor model.
            Parameters are ordered as follows:

                1. Three diffusion tensor's eigenvalues
                2. Three lines of the eigenvector matrix each containing the
                   first, second and third coordinates of the eigenvector
                3. Fifteen elements of the kurtosis tensor
                4. Twenty-One elements of the covariance tensor

        gtab1: dipy.core.gradients.GradientTable
            A GradientTable class instance for first DDE diffusion epoch
        gtab2: dipy.core.gradients.GradientTable
            A GradientTable class instance for second DDE diffusion epoch
        S0 : float or ndarray, optional
            The non diffusion-weighted signal in every voxel, or across all
            voxels.

        Returns
        -------
        S : numpy.ndarray
            Predicted signal based on the CTI model
        """

        return cti_prediction(cti_params, self.gtab1, self.gtab2, S0=S0)


class CorrelationTensorFit(DiffusionKurtosisFit):
    """Class for fitting the Correlation Tensor Model"""

    def __init__(self, model, model_params):
        """Initialize a CorrelationTensorFit class instance.

        Parameters
        ----------
        model : CorrelationTensorModel Class instance
            Class instance containing the Correlation Tensor Model for the fit
        model_params : ndarray (x, y, z, 48) or (n, 48)
            All parameters estimated from the diffusion kurtosis model.
            Parameters are ordered as follows:

                1) Three diffusion tensor's eigenvalues
                2) Three lines of the eigenvector matrix each containing the
                   first, second and third coordinates of the eigenvector
                3) Fifteen elements of the kurtosis tensor
                4) Twenty One elements of the covariance tensor

        """
        DiffusionKurtosisFit.__init__(self, model, model_params)

    @property
    def ct(self):
        """
        Returns the 21 independent elements of the covariance tensor as an
        array
        """
        return self.model_params[..., 27:48]

    @warning_for_keywords()
    def predict(self, gtab1, gtab2, *, S0=1):
        """Given a CTI model fit, predict the signal on the vertices of a
        gradient table

        Parameters
        ----------
        gtab1: dipy.core.gradients.GradientTable
            A GradientTable class instance for first DDE diffusion epoch
        gtab2: dipy.core.gradients.GradientTable
            A GradientTable class instance for second DDE diffusion epoch
        S0 : float or ndarray, optional
            The non diffusion-weighted signal in every voxel, or across all
            voxels. Default: 1

        Returns
        -------
        S : numpy.ndarray
            Predicted signal based on the CTI model
        """
        return cti_prediction(self.model_params, gtab1, gtab2, S0=S0)

    @property
    def K_aniso(self):
        r"""Returns the anisotropic Source of Kurtosis ($K_{aniso}$)

        Notes
        -----
        The $K_{aniso}$ is defined as :footcite:p:`NetoHenriques2020`:

        .. math::

            \[K_{aniso} = \frac{6}{5} \cdot \frac{\langle V_{\lambda}(D_c)
                                              \rangle}{\overline{D}^2}\]

        where $K_{aniso}$ is the anisotropic kurtosis,
        $\langle V_{\lambda}(D_c) \rangle$ represents the mean of the variance
        of eigenvalues of the diffusion tensor, $\overline{D}$ is the mean of
        the diffusion tensor.

        References
        ----------
        .. footbibliography::
        """
        C = self.ct
        D = self.quadratic_form

        Variance = (
            2
            / 9
            * (
                C[..., 0]
                + D[..., 0, 0] ** 2
                + C[..., 1]
                + D[..., 1, 1] ** 2
                + C[..., 2]
                + D[..., 2, 2] ** 2
                - C[..., 5]
                - D[..., 0, 0] * D[..., 1, 1]
                - C[..., 4]
                - D[..., 0, 0] * D[..., 2, 2]
                - C[..., 3]
                - D[..., 1, 1] * D[..., 2, 2]
                + 3
                * (
                    C[..., 17]
                    + D[..., 0, 1] ** 2
                    + C[..., 16]
                    + D[..., 0, 2] ** 2
                    + C[..., 15]
                    + D[..., 1, 2] ** 2
                )
            )
        )
        mean_D = np.trace(D) / 3
        if mean_D == 0:
            K_aniso = 0
        else:
            K_aniso = (6 / 5) * (Variance / (mean_D**2))

        return K_aniso

    @property
    def K_iso(self):
        r"""Returns the isotropic Source of Kurtosis ($K_{iso}$)

        Notes
        -----
        The $K_{iso}$ is defined as :

        .. math::

            K_{\text{iso}} = 3 \cdot \frac{V(\overline{D}^c)}{\overline{D}^2}

        where: $K_{\text{iso}}$ is the isotropic kurtosis, $V({\overline{D}^c})$
        represents the variance of the diffusion tensor raised to the power $c$,
        $\overline{D}$ is the mean of the diffusion tensor.

        """
        C = self.ct
        mean_D = self.md
        Variance = (
            1
            / 9
            * (
                C[..., 0]
                + C[..., 1]
                + C[..., 2]
                + 2 * C[..., 5]
                + 2 * C[..., 4]
                + 2 * C[..., 3]
            )
        )
        if mean_D == 0:
            K_iso = 0
        else:
            K_iso = 3 * (Variance / (mean_D**2))
        return K_iso

    @auto_attr
    def K_total(self):
        r""" Returns the total excess kurtosis.

            Notes
            -----
            $K_{total}$ is defined as:

            .. math::
                \[\Psi = \frac{2}{5} \cdot \frac{D_{11}^2 + D_{22}^2 + D_{33}^2
                                                 + 2D_{12}^2 + 2D_{13}^2 +
                                                 2D_{23}^2{\overline{D}^2} -
                                                 \frac{6}{5} \]
                \[{\overline{W}} = \frac{1}{5} \cdot (W_{1111} + W_{2222}
                                                      + W_{3333} + 2W_{1122}
                                                      + 2W_{1133} + 2W_{2233})\
                  ]

            where $\Psi$ is a variable representing a part of the total
            excess kurtosis, $D_{ij}$ are elements of the diffusion tensor,
            $\overline{D}$ is the mean of the diffusion tensor. $\overline{W}$
            is the mean kurtosis, $W_{ijkl}$ are elements of the kurtosis
            tensor.
        """

        mean_K = self.mkt()
        D = self.quadratic_form
        mean_D = self.md
        if mean_D == 0:
            psi = 0
        else:
            psi = 2 / 5 * (
                (
                    D[..., 0, 0] ** 2
                    + D[..., 1, 1] ** 2
                    + D[..., 2, 2] ** 2
                    + 2 * D[..., 0, 1] ** 2
                    + 2 * D[..., 0, 2] ** 2
                    + D[..., 1, 2] ** 2
                )
                / (mean_D**2)
            ) - (6 / 5)

        return mean_K + psi

    @property
    def K_micro(self):
        r"""Returns Microscopic Source of Kurtosis."""

        K_total = self.K_total
        K_aniso = self.K_aniso
        K_iso = self.K_iso
        micro_K = K_total - K_aniso - K_iso
        return micro_K


@warning_for_keywords()
def params_to_cti_params(result, *, min_diffusivity=0):
    # Extracting the diffusion tensor parameters from solution
    DT_elements = result[:6]
    evals, evecs = decompose_tensor(
        from_lower_triangular(DT_elements), min_diffusivity=min_diffusivity
    )
    # Extracting kurtosis tensor parameters from solution
    MD_square = evals.mean(0) ** 2
    KT_elements = result[6:21] / MD_square if MD_square else 0.0 * result[6:21]

    # Extracting correlation tensor parameters from solution
    CT_elements = result[21:42]

    # Write output
    cti_params = np.concatenate(
        (evals, evecs[0], evecs[1], evecs[2], KT_elements, CT_elements), axis=0
    )

    return cti_params


@warning_for_keywords()
def ls_fit_cti(
    design_matrix, data, inverse_design_matrix, *, weights=True, min_diffusivity=0
):
    r"""Compute the diffusion kurtosis and covariance tensors using an
    ordinary or weighted linear least squares approach

    Parameters
    ----------
    design_matrix : array (g, 43)
        Design matrix holding the covariants used to solve for the regression
        coefficients.
    data : array (g)
        Data or response variables holding the data.
    inverse_design_matrix : array (43, g)
        Inverse of the design matrix.
    weights : bool, optional
        Parameter indicating whether weights are used.
    min_diffusivity : float, optional
        Because negative eigenvalues are not physical and small eigenvalues,
        much smaller than the diffusion weighting, cause quite a lot of noise
        in metrics such as fa, diffusivity values smaller than `min_diffusivity`
        are replaced with `min_diffusivity`.

    Returns
    -------
    cti_params : array (48)
        All parameters estimated from the diffusion kurtosis model for all N
        voxels. Parameters are ordered as follows:

            1) Three diffusion tensor eigenvalues.
            2) Three blocks of three elements, containing the first second and
                third coordinates of the diffusion tensor eigenvectors.
            3) Fifteen elements of the kurtosis tensor.
            4) Twenty One elements of the covariance tensor.

    """
    A = design_matrix
    y = np.log(data)
    result = np.dot(inverse_design_matrix, y)
    if weights:
        W = np.diag(np.exp(2 * np.dot(A, result)))
        AT_W = np.dot(A.T, W)
        inv_AT_W_A = np.linalg.pinv(np.dot(AT_W, A))
        AT_W_LS = np.dot(AT_W, y)
        result = np.dot(inv_AT_W_A, AT_W_LS)
    cti_params = params_to_cti_params(result, min_diffusivity=min_diffusivity)

    return cti_params


common_fit_methods = {
    "WLS": ls_fit_cti,
    "OLS": ls_fit_cti,
    "UWLLS": ls_fit_cti,
    "ULLS": ls_fit_cti,
    "WLLS": ls_fit_cti,
    "OLLS": ls_fit_cti,
}
